from typing import Optional, Tuple

from ..expressions import (
    AddExpression,
    MathExpression,
    MultiplyExpression,
    SubtractExpression,
)
from ..rule import BaseRule, ExpressionChangeRule
from ..util import FactorResult, TermEx, factor_add_terms_ex, get_term_ex, make_term

_POS_SIMPLE = "simple"
_POS_CHAINED_BOTH = "chained_both"
_POS_CHAINED_LEFT = "chained_left"
_POS_CHAINED_LEFT_RIGHT = "chained_left_right"
_POS_CHAINED_RIGHT_LEFT = "chained_right_left"
_POS_CHAINED_RIGHT = "chained_right"


class DistributiveFactorOutRule(BaseRule):
    r"""Distributive Property
        `ab + ac = a(b + c)`

         The distributive property can be used to expand out expressions
         to allow for simplification, as well as to factor out common properties
         of terms.

         **Factor out a common term**

         This handles the `ab + ac` conversion of the distributive property, which
         factors out a common term from the given two addition operands.

                   +               *
                  / \             / \
                 /   \           /   \
                /     \    ->   /     \
               *       *       a       +
              / \     / \             / \
             a   b   a   c           b   c
    """
    constants: bool

    def __init__(self, constants: bool = False):
        # If true, will factor common numbers out of a const+const expression
        self.constants = constants

    @property
    def name(self) -> str:
        return "Distributive Factoring"

    @property
    def code(self) -> str:
        return "DF"

    def get_type(self, node: MathExpression) -> Optional[Tuple[str, TermEx, TermEx]]:
        """Determine the configuration of the tree for this transformation.

        Support the three types of tree configurations:
         - Simple is where the node's left and right children are exactly
           terms linked by an add operation.
         - Chained Left is where the node's left child is a term, but the right
           child is another add operation. In this case the left child
           of the next add node is the target.
         - Chained Right is where the node's right child is a term, but the left
           child is another add operation. In this case the right child
           of the child add node is the target.

        Structure:
         - Simple
            * node(add),node.left(term),node.right(term)
         - Chained Left
            * node(add),node.left(term),node.right(add),node.right.left(term)
         - Chained Right
            * node(add),node.right(term),node.left(add),node.left.right(term)
        """
        if not isinstance(node, AddExpression):
            return None
        # Left node in both cases is always resolved as a term.
        left_term = get_term_ex(node.left)
        right_term = get_term_ex(node.right)

        # No terms found for either child
        if left_term is None and right_term is None:
            # Check for the rare case where both terms are chained.
            # This happens when forcing the associative groups into
            # a certain form. It's not usually useful, but it's a
            # valid thing to do.

            if isinstance(node.right, AddExpression):
                right_term = get_term_ex(node.right.left)
            if right_term is None or right_term.variable is None:
                return None

            if isinstance(node.left, AddExpression):
                left_term = get_term_ex(node.left.right)
            if left_term is None or left_term.variable is None:
                return None
            return _POS_CHAINED_BOTH, left_term, right_term

        # Simplest case of each child being a term exactly.
        if left_term is not None and right_term is not None:
            return _POS_SIMPLE, left_term, right_term

        # Left child is a term
        if left_term is not None:
            # TODO: I'm not sure why I had this restriction here.
            # TODO: add a comment about it when you remember.
            if left_term.variable is None:
                return None
            if isinstance(node.right, AddExpression):
                right_term = get_term_ex(node.right.left)
            if right_term is not None:
                if right_term.variable is None:
                    return None
                return (
                    _POS_CHAINED_RIGHT,
                    left_term,
                    right_term,
                )

            # check inside another group
            if isinstance(node.right, AddExpression) and isinstance(
                node.right.left, AddExpression
            ):
                right_term = get_term_ex(node.right.left.left)
            if right_term is None or right_term.variable is None:
                return None
            return (
                _POS_CHAINED_RIGHT_LEFT,
                left_term,
                right_term,
            )

        # Right child is a term
        if right_term is not None:
            # TODO: I'm not sure why I had this restriction here.
            # TODO: add a comment about it when you remember.
            if right_term.variable is None:
                return None
            if isinstance(node.left, AddExpression):
                left_term = get_term_ex(node.left.right)
            if left_term is not None:
                if left_term.variable is None:
                    return None
                return _POS_CHAINED_LEFT, left_term, right_term

            # check inside another group
            if isinstance(node.left, AddExpression) and isinstance(
                node.left.right, AddExpression
            ):
                left_term = get_term_ex(node.left.right.right)
            if left_term is None or left_term.variable is None:
                return None
            return (
                _POS_CHAINED_LEFT_RIGHT,
                left_term,
                right_term,
            )

        return None

    def can_apply_to(self, node: MathExpression) -> bool:
        type_tuple = self.get_type(node)
        if type_tuple is None:
            return False
        type, l_term, r_term = type_tuple
        # Don't try factoring out terms with no variables, e.g "4 + 84"
        if (
            self.constants is False
            and l_term.variable is None
            and r_term.variable is None
        ):
            return False

        f = factor_add_terms_ex(l_term, r_term)
        if not f:
            return False

        if f.best == 1 and not f.variable and not f.exponent:
            return False

        return True

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node).save_parent()
        type_result = self.get_type(node)
        assert (
            type_result is not None
        ), "make sure can_apply_to returns True before calling apply_to"
        tree_position, left_term, right_term = type_result
        assert left_term is not None
        assert right_term is not None
        factors = factor_add_terms_ex(left_term, right_term)
        assert isinstance(factors, FactorResult)
        a = make_term(factors.best, factors.variable, factors.exponent)
        b = make_term(factors.left, factors.leftVariable, factors.leftExponent)
        c = make_term(factors.right, factors.rightVariable, factors.rightExponent)
        inside = (
            AddExpression(b, c)
            if isinstance(node, AddExpression)
            else SubtractExpression(b, c)
        )
        # NOTE: we swap the output order of the extracted
        #       common factor and what remains to prefer
        #       ordering that can be expressed without an
        #       explicit multiplication symbol.
        result: MathExpression = MultiplyExpression(inside, a)
        result.all_changed()

        # Fix the links to existing nodes on the left side of the result
        left_positions = [
            _POS_CHAINED_LEFT,
            _POS_CHAINED_BOTH,
        ]
        if tree_position in left_positions:
            assert node.left is not None
            # Because in the chained mode we extract node.left.right, the other
            # child is the remainder we want to be sure to preserve.
            # e.g. "(4 + p) + p" we need to keep "4"
            keep_child = node.left.left
            result = AddExpression(keep_child, result)

        # Fix the links to existing nodes on the left-right side of the result
        if tree_position == _POS_CHAINED_LEFT_RIGHT:
            assert node.left is not None and node.left.right is not None
            keep_child = AddExpression(node.left.left, node.left.right.left)
            result = AddExpression(keep_child, result)

        # Fix the links to existing nodes on the right-left side of the result
        if tree_position == _POS_CHAINED_RIGHT_LEFT:
            assert node.right is not None and node.right.left is not None
            keep_child = AddExpression(node.right.left.right, node.right.right)
            result = AddExpression(result, keep_child)

        # Fix the links to existing nodes on the right side of the result
        right_positions = [
            _POS_CHAINED_RIGHT,
            _POS_CHAINED_BOTH,
        ]
        if tree_position in right_positions:
            assert node.right is not None
            # Because in the chained mode we extract node.right.left, the other
            # child is the remainder we want to be sure to preserve.
            # e.g. "p + (p + 2x)" we need to keep 2x
            keep_child = node.right.right
            result = AddExpression(result, keep_child)
        change.done(result)
        return change



from typing import List

import numpy as np
import pytest

from mathy_core.tree import STOP, BinaryTreeNode

from .conftest import BinarySearchTree


def test_tree_node_constructor():
    """verify that the children passed in the constructor are properly assigned,
    and that their `parent` variables are also set to the root node properly."""

    tree = BinaryTreeNode(BinaryTreeNode(), BinaryTreeNode())
    assert tree.left is not None and tree.right is not None
    count = 0

    def node_visit(node, depth, data):
        nonlocal count
        count = count + 1

    tree.visit_inorder(node_visit)
    assert count == 3
    assert tree.left.parent == tree
    assert tree.right.parent == tree


def test_tree_node_clone():
    """check to be sure that when we clone off a known node of the tree
    that its children, and only its children, are still searchable
    from the cloned tree."""
    tree = BinarySearchTree(0)
    for i in range(26):
        tree.insert(i)
    fifteen = tree.find(15)
    assert fifteen is not None
    clone = fifteen.clone()
    # The clone node becomes the root of the new tree
    assert clone.get_root() == clone
    count = 0

    def node_visit(node, depth, data):
        nonlocal count
        count = count + 1

    clone.visit_inorder(node_visit)
    # clone node has expected 11 remaining nodes
    assert count == 11


def test_tree_node_is_leaf():

    """check that the known extremes of a tree are reported as leaf nodes
    and that all other known non - extremes are not."""
    tree = BinarySearchTree(0)
    for i in range(-1, 6):
        tree.insert(i)
    assert tree.find(-1).is_leaf() is True
    assert tree.find(5).is_leaf() is True
    for i in range(4):
        assert tree.find(i).is_leaf() is False


def test_tree_node_rotate():
    """test to ensure that rotations do not compromise the search tree
    by randomly rotating nodes and verifying that all known numbers can
    still be found."""
    tree = BinarySearchTree(0)
    values = []
    for i in range(-5, 6):
        values.append(i)
        tree.insert(i)
    for i in range(1000):
        index = np.random.randint(0, len(values))  # type:ignore
        node = tree.find(values[index])
        node.rotate()
    for v in values:
        assert tree.find(v) is not None


def test_tree_node_visit_stop():
    """Verify that tree visits can be stopped by returning the STOP constant"""

    values = [-1, 0, 1]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)
    total = 0

    def visit_pre(node, depth, data):
        nonlocal total
        total += 1
        if node.key == -1:
            return STOP

    tree.visit_preorder(visit_pre)
    # preorder stops at second node
    assert total == 2

    total = 0

    def visit_in(node, depth, data):
        nonlocal total
        total += 1
        if node.key == -1:
            return STOP

    tree.visit_inorder(visit_in)
    # inorder stops at first node
    assert total == 1

    total = 0

    def visit_post(node, depth, data):
        nonlocal total
        total += 1
        if node.key == -1:
            return STOP

    tree.visit_postorder(visit_post)
    # postorder stops at first node
    assert total == 1


def test_tree_node_visit_preorder():
    values = [-1, 0, 1]
    order = [0, -1, 1]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)

    def node_visit(node, depth, data):
        nonlocal order
        assert node.key == order.pop(0)

    tree.visit_preorder(node_visit)


def test_tree_node_visit_inorder():
    values = [-1, 0, 1]
    order = [-1, 0, 1]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)

    def node_visit(node, depth, data):
        nonlocal order
        assert node.key == order.pop(0)

    tree.visit_inorder(node_visit)


def test_tree_node_visit_postorder():
    values = [-1, 0, 1]
    order = [-1, 1, 0]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)

    def node_visit(node, depth, data):
        nonlocal order
        assert node.key == order.pop(0)

    tree.visit_postorder(node_visit)


def test_tree_node_get_root():
    tree = BinarySearchTree(0)
    values = list(range(-5, 6))
    for i in values:
        tree.insert(i)
    for i in values:
        assert tree.find(i).get_root() == tree


def test_tree_node_set_left():
    one = BinaryTreeNode()
    two = BinaryTreeNode()
    three = BinaryTreeNode()
    one.set_left(two)
    assert one.left == two
    assert two.parent == one
    with pytest.raises(ValueError):
        # Cannot set self to child
        one.set_right(one)

    # can clear `.parent` on a child that is replaced
    assert two.parent == one
    one.set_left(three, clear_old_child_parent=True)
    assert two.parent is None


def test_tree_node_set_right():
    one = BinaryTreeNode()
    two = BinaryTreeNode()
    three = BinaryTreeNode()
    one.set_right(two)
    assert one.right == two
    assert two.parent == one

    with pytest.raises(ValueError):
        # Cannot set self to child
        one.set_right(one)

    # can clear `.parent` on a child that is replaced
    assert two.parent == one
    one.set_right(three, clear_old_child_parent=True)
    assert two.parent is None


def test_tree_node_get_side():
    values = list(range(-4, 0)) + list(range(1, 5))
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)
    node = tree.find(-4)
    assert node is not None
    assert node.parent.get_side(node) == "left"
    node = tree.find(4)
    assert node is not None
    assert node.parent.get_side(node) == "right"
    with pytest.raises(ValueError):
        # Raises an error if the child does not belong to this parent
        node.parent.get_side(BinaryTreeNode())


def test_tree_node_set_side():
    tree = BinaryTreeNode()
    one = BinaryTreeNode()
    two = BinaryTreeNode()
    tree.set_side(one, "left")
    assert tree.left == one
    tree.set_side(two, "right")
    assert tree.right == two
    with pytest.raises(ValueError):
        # error if side name is not known
        tree.set_side(one, "rihgt")  # type:ignore


def test_tree_node_get_children():
    values = [-2, -1, -3, 0, 1, 2]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)
    neg: List[BinarySearchTree] = tree.find(-2).get_children()
    assert len(neg) == 2
    assert neg[0].key == -3
    assert neg[1].key == -1
    one: List[BinarySearchTree] = tree.find(1).get_children()
    assert len(one) == 1
    assert one[0].key == 2
    two = tree.find(2).get_children()
    assert len(two) == 0


def test_tree_node_get_sibling():
    tree = BinaryTreeNode(BinaryTreeNode(), BinaryTreeNode())
    assert tree.left.get_sibling() == tree.right
    assert tree.right.get_sibling() == tree.left
    assert tree.get_sibling() is None



from typing import Any, Optional

from mathy_core.tree import BinaryTreeNode


class BinarySearchTree(BinaryTreeNode):
    left: Optional["BinarySearchTree"]
    right: Optional["BinarySearchTree"]
    parent: Optional["BinarySearchTree"]

    def __init__(self, key: int = None, **kwargs: Any):
        super(BinarySearchTree, self).__init__(**kwargs)
        self.key = key
        self.x = 0
        self.y = 0
        self.offset = 0

    def clone(self) -> "BinarySearchTree":  # type:ignore[override]
        result: BinarySearchTree = super(BinarySearchTree, self).clone()
        result.key = self.key
        return result

    def insert(self, key: int) -> BinaryTreeNode:
        """Insert a node in the tree with the specified key."""
        node: BinarySearchTree = self.get_root()
        while node:
            assert node.key is not None
            if key > node.key:
                if not node.right:
                    node.set_right(BinarySearchTree(key))
                    break
                node = node.right
            elif key < node.key:
                if not node.left:
                    node.set_left(BinarySearchTree(key))
                    break

                node = node.left
            else:
                break
        return self

    def find(self, key: int) -> Optional[BinaryTreeNode]:
        """Find a node in the tree by its key and return it.  Return None if the key
        is not found in the tree."""
        node = self.get_root()
        while node:
            assert node.key is not None
            if key > node.key:
                if not node.right:
                    return None

                node = node.right
                continue

            if key < node.key:
                if not node.left:
                    return None

                node = node.left
                continue

            if key == node.key:
                return node

            return None

        return None



import math
import random
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union, cast

import numpy as np
from wasabi import TracebackPrinter

from .expressions import (
    AddExpression,
    BinaryExpression,
    ConstantExpression,
    DivideExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
    PowerExpression,
    SubtractExpression,
    VariableExpression,
)
from .parser import ExpressionParser
from .tree import LEFT, VisitDataType, VisitStop
from .types import Literal, NumberType


def is_debug_mode() -> bool:
    """Debug mode enables extra logging and assertions, but is slower."""
    return False


def compare_expression_string_values(
    from_expression: str, to_expression: str, history: Optional[List[Any]] = None
) -> None:
    """Compare and evaluate two expressions strings to verify they have the
    same value"""
    parser = ExpressionParser()
    return compare_expression_values(
        parser.parse(from_expression), parser.parse(to_expression), history
    )


def raise_with_history(
    title: str,
    description: str,
    history: Optional[List[Any]] = None,
) -> None:
    import traceback

    history_text: List[str] = []
    if history is not None:
        history_text = [f" - {h.raw}" for h in history]
    history_text.insert(0, description)
    tb = TracebackPrinter()
    error = tb(title, "\n".join(history_text), tb=traceback.extract_stack())
    raise ValueError(error)


def compare_expression_values(
    from_expression: MathExpression,
    to_expression: MathExpression,
    history: Optional[List[Any]] = None,
) -> None:
    """Compare and evaluate two expressions to verify they have the same value"""
    vars_from: set = set()
    vars_to: set = set()
    for v in from_expression.find_type(VariableExpression):
        vars_from.add(v.identifier)
    for v in to_expression.find_type(VariableExpression):
        vars_to.add(v.identifier)
    # If there are not the same unique vars in the two expressions, something
    # bad happened, and the two expressions can only coincidentally be equal
    # in value.
    if len(vars_from) != len(vars_to):
        raise_with_history(
            "Number of variables changed",
            f"{list(vars_from)} != {list(vars_to)}",
            history,
        )

    sorted_from = list(vars_from)
    sorted_from.sort()
    sorted_to = list(vars_from)
    sorted_to.sort()

    if sorted_from != sorted_to:
        raise_with_history(
            "Unique variables changed",
            f"{sorted_from} != {sorted_to}",
            history,
        )

    # Generate random values for each variable, and then evaluate the expressions
    eval_context: Dict[str, NumberType] = {}
    for var in vars_from:
        eval_context[var] = float(random.randint(1, 100))

    value_from = from_expression.evaluate(eval_context)
    value_to = to_expression.evaluate(eval_context)

    # Print out the problem steps leading up to error result.
    if not math.isclose(value_from, value_to, rel_tol=1e-9, abs_tol=0.0):
        changed = f"""
        {from_expression} = {value_from}

        {to_expression} = {value_to}

        {value_from} != {value_to}
        """
        raise_with_history("Expression value changed", changed, history)


def unlink(node: Optional[MathExpression] = None) -> Optional[MathExpression]:
    """Unlink an expression from it's parent.

    1. Clear expression references in `parent`
    2. Clear `parent` in expression
    """

    if node is None:
        return None

    if node.parent is not None and node == node.parent.left:
        node.parent.set_left(None)

    if node.parent is not None and node == node.parent.right:
        node.parent.set_right(None)

    node.parent = None
    return node


# Build a verbose factor dictionary.
#
# This builds a dictionary of factors for a given value that
# contains both arrangements of terms so that all factors are
# accessible by key.  That is, factoring 2 would return
#      result =
#        1 : 2
#        2 : 1
def factor(value: NumberType) -> Dict[NumberType, NumberType]:
    if value == 0 or math.isnan(value):
        return {}
    np.seterr(invalid="ignore")  # type:ignore
    sqrt = np.sqrt(value)  # type:ignore
    np.seterr(invalid="warn")  # type:ignore
    if math.isnan(sqrt):
        return {1: value}

    sqrt = int(sqrt + 1)
    factors: Dict[NumberType, NumberType] = {1: value}
    factors[value] = 1

    for i in range(2, sqrt):
        if value % i == 0:
            one = i
            two = value / i
            factors[one] = two
            factors[two] = one
    return factors


def is_add_or_sub(node: MathExpression) -> bool:
    return isinstance(node, AddExpression) or isinstance(node, SubtractExpression)


def get_sub_terms(
    node: MathExpression,
) -> Union[
    Literal[False],
    List[
        Tuple[
            Optional[ConstantExpression],
            Optional[VariableExpression],
            Optional[PowerExpression],
        ]
    ],
]:
    nodes = node.to_list("inorder")
    terms = []

    def safe_pop() -> Optional[MathExpression]:
        nonlocal nodes
        if len(nodes) > 0:
            return nodes.pop(0)
        return None

    current = safe_pop()

    while current is not None:
        term_const = term_var = term_exp = None
        # If there's a coefficient, note it
        if isinstance(current, ConstantExpression):
            term_const = current
            current = safe_pop()
            # Cannot have add/sub in sub-terms
            if current is not None and is_add_or_sub(current):
                return False
            # It should be one of these operators
            assert current is None or isinstance(
                current, (MultiplyExpression, DivideExpression, PowerExpression)
            )
            if not isinstance(current, PowerExpression):
                current = safe_pop()

        # Pop off the variable
        if isinstance(current, VariableExpression):
            term_var = current
            current = safe_pop()
            # cannot have add/sub in sub-terms
            if current is not None and is_add_or_sub(current):
                return False
            # It should be one of these operators
            assert current is None or isinstance(
                current, (MultiplyExpression, DivideExpression, PowerExpression)
            )
            if not isinstance(current, PowerExpression):
                current = safe_pop()

        # Look for exponent
        if isinstance(current, PowerExpression):
            # pop the power binary expression
            current = safe_pop()
            # store the right hand side
            term_exp = current
            # pop it off for next term
            current = safe_pop()

        # Couldn't find anything
        if term_const is None and term_exp is None and term_var is None:
            # If it's a continuation multiply, e.g in "4 * (x^2 * z^6)"
            # NOTE: it's a continuation mult in the above expression
            #   because of the parens grouping between the 4 multiply
            #   and the inner multiply of the parens sub-terms
            if isinstance(current, MultiplyExpression):
                # pop it off for next term
                current = safe_pop()
                continue

            return False
        terms.append((term_const, term_var, term_exp))
    return terms  # type: ignore


def is_simple_term(node: MathExpression) -> bool:
    """Return True if a given term has been simplified so it only has at
    most one of each variable and a constant.

    # Examples
      - Simple = 2x^2 * 2y
      - Complex = 2x * 2x * 2y
      - Simple = x^2 * 4
      - Complex = 2 * 2x^2
    """
    sub_terms = get_sub_terms(node)
    if sub_terms is False:
        return False
    assert isinstance(sub_terms, list)
    seen: set = set()
    co_key = "coefficient"

    for coefficient, variable, exponent in sub_terms:
        if coefficient is not None:
            if co_key in seen:
                return False
            seen.add(co_key)

        if variable is not None or exponent is not None:
            key = "{}{}".format(variable, exponent)
            if key in seen:
                return False
            seen.add(key)

    return True


def is_preferred_term_form(expression: MathExpression) -> bool:
    """
    Return True if a given term has been simplified so that it only has
    a max of one coefficient and variable, with the variable on the right
    and the coefficient on the left side

    Examples

      - Complex   = 2 * 2x^2
      - Simple    = x^2 * 4
      - Preferred = 4x^2
    """
    if not is_simple_term(expression):
        return False

    # If there are multiple multiplications this term can be simplified further. At most
    # we expect a multiply to connect a coefficient and variable.
    # NOTE: the following check is removed because we need to handle multiple variable
    #       terms, e.g. "4x * z"
    # if len(expression.find_type(MultiplyExpression)) > 1:
    #     return False

    # if there's a variable, make sure the coefficient is on the left side
    # for the preferred compact form. i.e. "4x" instead of "x * 4"
    vars: List[VariableExpression] = expression.find_type(VariableExpression)
    seen_vars: Dict[str, int] = dict()
    parent: MathExpression
    for var in vars:
        if var.identifier is None:
            continue
        if var.identifier not in seen_vars:
            seen_vars[var.identifier] = 0
        seen_vars[var.identifier] += 1
        parent = var
        if isinstance(var.parent, PowerExpression):
            parent = var.parent
        if parent.parent is not None and parent.parent.get_side(parent) == LEFT:
            if isinstance(parent.parent.right, ConstantExpression):
                return False
    # If any variables show up more than once, they can be combined.
    # examples:
    #  - `b * (44b^2)`
    #  - `z * (1274z^2)`
    for key, value in seen_vars.items():
        if value > 1:
            return False
    return True


def has_like_terms(expression: MathExpression) -> bool:
    """Return True if a given expression has more than one of any type of term.

    # Examples

    - `x + y + z` = `False`
    - `x^2 + x` = `False`
    - `y + 2x` = `True`
    - `x^2 + 4x^3 + 2y` = `True`
    """

    seen: set = set()
    term_nodes = get_terms(expression)
    for node in term_nodes:
        term = get_term(node)
        if not isinstance(term, TermResult):
            continue
        var_key = ("".join(term.variables), term.exponent)
        # If the same var/power combinaton is found in the expression more than once
        # there are like terms.
        if var_key in seen:
            return True
        seen.add(var_key)

    # Look for multiple free-floating constants
    consts = expression.find_type(ConstantExpression)
    for const in consts:
        if const.parent and is_add_or_sub(const.parent):
            if "const_term" in seen:
                return True
            seen.add("const_term")
    return False


class FactorResult:
    best: NumberType
    left: NumberType
    right: NumberType
    all_left: Dict[NumberType, NumberType]
    all_right: Dict[NumberType, NumberType]
    variable: Optional[str]
    exponent: Optional[NumberType]
    leftExponent: Optional[NumberType]
    rightExponent: Optional[NumberType]
    leftVariable: Optional[str]
    rightVariable: Optional[str]

    def __init__(self) -> None:
        self.best = -1
        self.left = -1
        self.right = -1
        self.all_left = {}
        self.all_right = {}
        self.variable = None
        self.exponent = None
        self.leftExponent = None
        self.rightExponent = None
        self.leftVariable = None
        self.rightVariable = None


# Create a term node hierarchy from a given set of
# term parameters.  This takes into account removing
# implicit coefficients of 1 where possible.
def make_term(
    coefficient: NumberType = 1,
    variable: Optional[str] = None,
    exponent: Optional[NumberType] = None,
) -> MathExpression:
    constExp = ConstantExpression(coefficient)
    if variable is None and exponent is None:
        return constExp

    varExp = VariableExpression(variable)
    if coefficient == 1 and exponent is None:
        return varExp

    multExp = MultiplyExpression(constExp, varExp)
    if exponent is None:
        return multExp

    expConstExp = ConstantExpression(exponent)
    if coefficient == 1:
        return PowerExpression(varExp, expConstExp)

    return PowerExpression(multExp, expConstExp)


class TermResult:
    coefficients: List[NumberType]
    variables: List[str]
    exponent: Optional[NumberType]
    node_coefficients: List[ConstantExpression]
    node_variables: List[VariableExpression]
    node_exponent: Optional[PowerExpression]

    def __init__(self) -> None:
        self.coefficients = []
        self.variables = []
        self.exponent = None
        self.node_coefficients = []
        self.node_variables = []
        self.node_exponent = None


# Extract term information from the given node
#
def get_term(node: MathExpression) -> Union[TermResult, Literal[False]]:
    result = TermResult()
    # Constant with add/sub parent should be OKAY.
    if isinstance(node, ConstantExpression):
        if not node.parent or (node.parent and is_add_or_sub(node.parent)):
            assert node.value is not None
            result.coefficients = [node.value]
            result.node_coefficients = [node]
            return result

    # Variable with add/sub parent should be OKAY.
    if isinstance(node, VariableExpression):
        if not node.parent or (node.parent and is_add_or_sub(node.parent)):
            assert node.identifier is not None
            result.variables = [node.identifier]
            result.node_variables = [node]
            return result

    # TODO: Comment resolution on whether +- is OKAY, and if not, why it breaks down.
    if not is_add_or_sub(node):
        if (
            len(node.find_type(AddExpression)) > 0
            or len(node.find_type(SubtractExpression)) > 0
        ):
            return False

    # If another add is found on the left side of this node, and the right node
    # is _NOT_ a leaf, we cannot extract a term.  If it is a leaf, the term should be
    # just the right node.
    if node.left and len(node.left.find_type(AddExpression)) > 0:
        if node.right and not node.right.is_leaf():
            return False

    if node.right and len(node.right.find_type(AddExpression)) > 0:
        return False

    exponents = node.find_type(PowerExpression)
    if len(exponents) > 0:
        # Supports only single exponents in terms
        if len(exponents) != 1:
            return False

        exponent = exponents[0]
        if not isinstance(exponent.right, ConstantExpression):
            return False

        result.exponent = exponent.right.value
        result.node_exponent = exponent

    variables = node.find_type(VariableExpression)
    if len(variables) > 0:
        result.variables = [v.identifier for v in variables if v.identifier is not None]
        result.variables.sort()
        result.node_variables = variables

    def filter_coefficients(n: MathExpression) -> bool:
        if not n.parent or n.parent == node.parent:
            return True

        if isinstance(n.parent, BinaryExpression) and not isinstance(
            n.parent, MultiplyExpression
        ):
            return False
        return True

    coefficients = node.find_type(ConstantExpression)
    coefficients = [c for c in coefficients if filter_coefficients(c)]
    if len(coefficients) > 0:

        def resolve_coefficients(c: ConstantExpression) -> NumberType:
            value = c.value
            assert value is not None
            if isinstance(c.parent, NegateExpression):
                value *= -1

            return value

        result.coefficients = [resolve_coefficients(c) for c in coefficients]
        result.node_coefficients = coefficients

    empty = (
        len(result.variables) == 0
        and len(result.coefficients) == 0
        and result.exponent is None
    )
    if empty:
        return False

    # consistently return an empty coefficients/variables array in case none exist.
    # this ensures that you can always reference coefficients[0] or variables[0] and
    # check that for truthiness, rather than having to check that the object property
    # `coefficients` or `variables` is not None and also the truthiness of index 0.
    if not result.coefficients:
        result.coefficients = []

    if not result.variables:
        result.variables = []
    return result


class TermEx(NamedTuple):
    coefficient: Optional[NumberType]
    variable: Optional[str]
    exponent: Optional[NumberType]


# fmt: off
TermEx.coefficient.__doc__ = "An optional integer or float coefficient" # noqa
TermEx.variable.__doc__ = "An optional variable" # noqa
TermEx.exponent.__doc__ = "An optional integer or float exponent" # noqa
# fmt: on


def get_term_ex(node: Optional[MathExpression]) -> Optional[TermEx]:
    """Extract the 3 components of a naturally ordered term.

    !!! info Important

        This doesn't care about whether the node is part of a larger term,
        it only looks at its children.

    # Example

    `mathy:4x^7`

    ```python
    TermEx(coefficient=4, variable="x", exponent=7)
    ```
    """
    if isinstance(node, NegateExpression):
        child = cast(NegateExpression, node).get_child()

        # "-x"
        if isinstance(child, VariableExpression):
            return TermEx(-1, child.identifier, None)
        # "-x^2"
        if isinstance(child, PowerExpression):
            if isinstance(child.left, VariableExpression) and isinstance(
                child.right, ConstantExpression
            ):
                return TermEx(-1, child.left.identifier, child.right.value)

    # "4"
    if isinstance(node, ConstantExpression):
        # Make sure the parent isn't an exponent link (in which case we'd incorrectly
        # report this as a term with no exponent.) That case can be handled by calling
        # this on the parent node.
        if node.parent is None or not isinstance(node.parent, PowerExpression):
            return TermEx(node.value, None, None)

    # "x"
    if isinstance(node, VariableExpression):
        # Make sure the parent isn't an exponent link (in which case we'd incorrectly
        # report this as a term with no exponent.) That case can be handled by calling
        # this on the parent node.
        if node.parent is None or not isinstance(node.parent, PowerExpression):
            return TermEx(None, node.identifier, None)
    # "4 * ???"
    if isinstance(node, MultiplyExpression) and isinstance(
        node.left, ConstantExpression
    ):
        # "4x"
        if isinstance(node.right, VariableExpression):
            return TermEx(node.left.value, node.right.identifier, None)

        # "4x^2"
        if isinstance(node.right, PowerExpression):
            pow = node.right
            if isinstance(pow.left, VariableExpression) and isinstance(
                pow.right, ConstantExpression
            ):
                return TermEx(node.left.value, pow.left.identifier, pow.right.value)

    # "x^2"
    if isinstance(node, PowerExpression):
        if isinstance(node.left, VariableExpression) and isinstance(
            node.right, ConstantExpression
        ):
            return TermEx(None, node.left.identifier, node.right.value)

    return None


def factor_add_terms_ex(
    left_term: TermEx, right_term: TermEx
) -> Union[FactorResult, Literal[False]]:
    if not left_term or not right_term:
        raise ValueError("invalid terms for factoring")

    # Common coefficients
    l_factors = factor(
        left_term.coefficient if left_term.coefficient is not None else 1
    )
    r_factors = factor(
        right_term.coefficient if right_term.coefficient is not None else 1
    )
    common = [k for k in r_factors if k in l_factors]
    if len(common) == 0:
        return False
    has_left: bool = left_term.variable is not None
    has_right: bool = right_term.variable is not None

    # If there are variables, we want to extract them, so
    # the smallest number to factor out. TODO: is this okay?
    if has_left or has_right:
        best = np.min(common)
    else:
        best = np.max(common)
    result = FactorResult()
    result.best = best
    result.left = l_factors[best]
    result.right = r_factors[best]
    result.all_left = l_factors
    result.all_right = r_factors

    # Common variables and powers
    two_exp_and_match: bool = (
        left_term.exponent is not None
        and right_term.exponent is not None
        and left_term.exponent == right_term.exponent
    )
    both_match = (
        False
        if (left_term.exponent or right_term.exponent) and not two_exp_and_match
        else True
    )
    if (
        has_left
        and has_right
        and left_term.variable == right_term.variable
        and both_match
    ):
        result.variable = left_term.variable
        result.exponent = left_term.exponent

    if left_term.exponent and left_term.exponent != result.exponent:
        result.leftExponent = left_term.exponent

    if right_term.exponent and right_term.exponent != result.exponent:
        result.rightExponent = right_term.exponent

    if has_left and left_term.variable != result.variable:
        result.leftVariable = left_term.variable

    if has_right and right_term.variable != result.variable:
        result.rightVariable = right_term.variable

    return result


def get_terms(expression: MathExpression) -> List[MathExpression]:
    """Walk the given expression tree and return a list of nodes
    representing the distinct terms it contains.

    # Arguments
    expression (MathExpression): the expression to find term nodes in

    # Returns
    (List[MathExpression]): a list of term nodes
    """
    results: List[MathExpression] = []
    root = expression.get_root()
    if isinstance(root, MultiplyExpression):
        results.append(root)

    def visit_fn(
        node: MathExpression, depth: int, data: VisitDataType
    ) -> Optional[VisitStop]:
        nonlocal results
        if not is_add_or_sub(node):
            return None
        if node.left and not is_add_or_sub(node.left):
            results.append(node.left)
        if node.right and not is_add_or_sub(node.right):
            results.append(node.right)
        return None

    root.visit_inorder(visit_fn)
    return [expression] if len(results) == 0 else results


def terms_are_like(
    one: Union[TermResult, MathExpression, Literal[False]],
    two: Union[TermResult, MathExpression, Literal[False]],
) -> bool:
    """Determine if two math expression nodes are **like terms**.

    # Arguments
    one (MathExpression): A math expression that represents a term
    two (MathExpression): Another math expression that represents a term

    # Returns
    (bool): Whether the terms are like or not.
    """
    # Extract terms from MathExpressions if need be
    if isinstance(one, MathExpression):
        one = get_term(one)

    if isinstance(two, MathExpression):
        two = get_term(two)

    if one is False or two is False:
        return False

    # Coerce the type away from math expression
    assert isinstance(one, TermResult) and isinstance(two, TermResult)

    # if neither have variables, then they are a match!
    if len(one.variables) == 0 and len(two.variables) == 0:
        return True

    # Both must have variables, and they must match exactly.
    if len(one.variables) != len(two.variables):
        return False

    invalid = len([False for v in one.variables if v not in two.variables]) > 0
    if invalid:
        return False

    # Also, the exponents must match
    if one.exponent != two.exponent:
        return False

    # Same variables, and exponents.  Good to go.
    return True


def pad_array(in_list: List[Any], max_length: int, value: Any = 0) -> List[Any]:
    """Pad a list to the given size with the given padding value.

    # Arguments:
    in_list (List[Any]): List of values to pad to the given length
    max_length (int): The desired length of the array
    value (Any): a value to insert in order to pad the array to max length

    # Returns
    (List[Any]): An array padded to `max_length` size
    """
    while len(in_list) < max_length:
        in_list.append(value)
    return in_list


def print_error(error: BaseException, text: str, print_error: bool = True) -> None:
    import traceback

    caught_error = TracebackPrinter(
        color_error="yellow", tb_base=".", tb_range_start=-15
    )(
        error.__class__.__name__,
        f"{error}",
        tb=traceback.extract_tb(error.__traceback__),
    )
    caught_at = TracebackPrinter(tb_base=".", tb_range_start=-15, tb_range_end=-1)(
        f"Error: {text}",
        "Caught at:",
        tb=traceback.extract_stack(),
    )

    if print_error:
        print(caught_at + caught_error)
    else:
        raise ValueError(caught_at + caught_error)



import sys
from typing import Union

# Use typing_extensions for Python < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Literal  # noqa
else:
    from typing import Literal  # noqa

NumberType = Union[float, int]



from ..expressions import AddExpression, MathExpression, MultiplyExpression
from ..rule import BaseRule, ExpressionChangeRule


class AssociativeSwapRule(BaseRule):
    r"""Associative Property
    Addition: `(a + b) + c = a + (b + c)`

             (y) +            + (x)
                / \          / \
               /   \        /   \
          (x) +     c  ->  a     + (y)
             / \                / \
            /   \              /   \
           a     b            b     c

     Multiplication: `(ab)c = a(bc)`

             (x) *            * (y)
                / \          / \
               /   \        /   \
          (y) *     c  <-  a     * (x)
             / \                / \
            /   \              /   \
           a     b            b     c
        """

    @property
    def name(self) -> str:
        return "Associative Group"

    @property
    def code(self) -> str:
        return "AG"

    def can_apply_to(self, node: MathExpression) -> bool:
        if isinstance(node.parent, AddExpression) and isinstance(node, AddExpression):
            return True

        if isinstance(node.parent, MultiplyExpression) and isinstance(
            node, MultiplyExpression
        ):
            return True

        return False

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node)
        node.rotate()
        node.set_changed()
        change.done(node)
        return change



import random
from typing import List, Optional

from mathy_core import MathExpression
from mathy_core.parser import ExpressionParser
from mathy_core.problems import (
    gen_simplify_multiple_terms,
    rand_number,
    use_pretty_numbers,
)
from mathy_core.util import TermEx, get_term_ex, get_terms


def test_number_generation() -> None:

    random.seed(1337)

    # When using pretty number generation, all values
    # are in range 1-12 and are always integers
    use_pretty_numbers(True)
    pretty_numbers = [rand_number() for _ in range(256)]
    outside_range_floats = [f for f in pretty_numbers if f < 1 or f > 12]
    pretty_floats = [f for f in pretty_numbers if isinstance(f, float)]
    assert len(outside_range_floats) == 0
    assert len(pretty_floats) == 0

    # When not using pretty numbers, values can be floats and large integers
    use_pretty_numbers(False)
    rand_numbers = [rand_number() for _ in range(256)]
    large_ints = [f for f in rand_numbers if isinstance(f, int) and f > 12]
    rand_floats = [f for f in rand_numbers if isinstance(f, float)]
    assert len(large_ints) > 0
    assert len(rand_floats) > 0


def test_problems_variable_sharing_like_terms() -> None:
    """Verify that the polynomial generation functions return matches that include
    shared variables with different exponents, e.g. "4x + x^3 + 2x + 1.3x^3"
    """
    parser = ExpressionParser()
    problem, _ = gen_simplify_multiple_terms(
        3, share_var_probability=1.0, noise_probability=0.0, op="+"
    )
    expression: MathExpression = parser.parse(problem)
    term_nodes: List[MathExpression] = get_terms(expression)
    found_var: Optional[str] = None
    found_exp: bool = False
    for term_node in term_nodes:
        ex: Optional[TermEx] = get_term_ex(term_node)
        assert ex is not None, f"invalid expression {term_node}"
        if found_var is None:
            found_var = ex.variable
        assert found_var == ex.variable, "expected only one variable"
        if ex.exponent is not None:
            found_exp = True

    # Assert there are terms with and without exponents for this var
    assert found_var is not None
    assert found_exp is True


def test_problems_variable_sharing_unlike_terms() -> None:
    """Verify that the polynomial generation functions return matches that include
    shared variables for terms that are not like, e.g. "4x + x^3 + 2x"
    """
    parser = ExpressionParser()
    problem, _ = gen_simplify_multiple_terms(
        2, share_var_probability=1.0, noise_probability=0.0, op="+"
    )
    expression: MathExpression = parser.parse(problem)
    term_nodes: List[MathExpression] = get_terms(expression)
    found_var: Optional[str] = None
    found_exp: bool = False
    for term_node in term_nodes:
        ex: Optional[TermEx] = get_term_ex(term_node)
        assert ex is not None, f"invalid expression {term_node}"
        if found_var is None:
            found_var = ex.variable
        assert found_var == ex.variable, "expected only one variable"
        if ex.exponent is not None:
            found_exp = True

    # Assert there are terms with and without exponents for this var
    assert found_var is not None
    assert found_exp is True


def test_problems_noise_grouping_terms() -> None:
    """Verify that the polynomial generation functions return matches that include
    non-standard grouping of terms, in order to force the agents to commute different
    tree configurations, rather than relying on say always commuting the 3rd node.

    Standard grouping is: "4x + x^3 + 2x + 7y"
    Grouped first two terms: "(4x + x^3) + 2x + 7y"
    Grouped second and third terms: "4x + (x^3 + 2x) + 7y"
    """
    for _ in range(100):
        problem, _ = gen_simplify_multiple_terms(6, grouping_noise_probability=1.0)
        assert "(" in problem and ")" in problem



__title__ = "mathy_core"
__version__ = "0.8.2"
__summary__ = "Computer Algebra System for working with math expressions"
__uri__ = "https://mathy.ai"
__author__ = "Justin DuJardin"
__email__ = "justin@dujardinconsulting.com"
__license__ = "All rights reserved"



"""Problem Generation
---

Utility functions for helping generate input problems.
"""
import random
from typing import Any, List, Optional, Set, Tuple, TypeVar, Union

from pydantic import BaseModel, Field

from .types import NumberType

DefaultType = TypeVar("DefaultType", bound=Any)
try:
    DefaultType.__doc__ = "Template type for a default return value"  # noqa
except AttributeError:
    ...  # Boo!

operators = list("+-*")
common_variables = list("xyz")
variables = list("abcdfghjklmnopqrstuvwxyz")
max_const = 12
_pretty_numbers = True


class MathyTermTemplate(BaseModel):
    variable: Optional[str] = Field(None, description="the term variable")
    exponent: Optional[int] = Field(None, description="the term exponent")

    def make(self) -> str:
        return mathy_term_string(
            coefficient=maybe_number(or_else=None),
            exponent=self.exponent,
            variable=self.variable,
        )


class MathyProblemTerm(MathyTermTemplate):
    coefficient: Optional[NumberType] = None


MathyExcludedTermTemplates = Optional[List[MathyTermTemplate]]


def mathy_term_string(
    *,
    coefficient: Optional[NumberType] = None,
    exponent: Optional[NumberType] = None,
    variable: Optional[str] = None,
) -> str:
    pieces = []
    if coefficient is not None:
        pieces.append(f"{coefficient}")
    if variable is not None:
        pieces.append(f"{variable}")
    if exponent is not None:
        pieces.append(f"^{exponent}")
    return "".join(pieces)


def get_rand_term_templates(
    num_templates: int,
    exclude_like: MathyExcludedTermTemplates = None,
    common_variables: bool = False,
    exponent_probability: float = 0.5,
) -> List[MathyTermTemplate]:
    result: List[MathyTermTemplate] = []

    exclude: List[str] = []
    if exclude_like is not None:
        exclude = [
            mathy_term_string(variable=t.variable, exponent=t.exponent)
            for t in exclude_like
        ]

    failures = 0
    while len(result) < num_templates:
        if failures > 100:
            raise EnvironmentError(
                f"failed to generate a random term after {failures} tries!"
            )
        variable = rand_var(common_variables)
        exponent: Optional[NumberType] = maybe_number(exponent_probability * 100, None)
        # Don't generate x^1
        if exponent == 1:
            exponent = 2
        key = mathy_term_string(variable=variable, exponent=exponent)
        if key not in exclude:
            result.append(MathyTermTemplate(variable=variable, exponent=exponent))
            exclude.append(key)
        else:
            failures += 1
    random.shuffle(result)
    return result


def rand_bool(percent_chance: NumberType = 50) -> bool:
    return bool(random.randrange(100) < percent_chance)


def rand_var(common: bool = False) -> str:
    if common is True:
        return common_variables[random.randint(0, len(common_variables) - 1)]
    return variables[random.randint(0, len(variables) - 1)]


def maybe_var(
    percent_chance: NumberType = 80,
    common_var: bool = False,
    or_else: DefaultType = "",  # type:ignore
) -> Union[DefaultType, str]:
    return rand_var(common_var) if rand_bool(percent_chance) else or_else


def maybe_number(
    percent_chance: NumberType = 80, or_else: DefaultType = ""  # type:ignore
) -> Union[int, float, DefaultType]:
    return rand_number() if rand_bool(percent_chance) else or_else


def maybe_power(
    percent_chance: NumberType = 80,
    max_power: int = 4,
    or_else: DefaultType = "",  # type:ignore
) -> Union[str, DefaultType]:
    if rand_bool(percent_chance):
        return "^{}".format(random.randint(2, max_power))
    else:
        return or_else


def use_pretty_numbers(enabled: bool = True) -> None:
    """Determine if problems should include only pretty numbers or
    a whole range of integers and floats. Using pretty numbers will
    restrict the numbers that are generated to integers between 1 and
    12. When not using pretty numbers, floats and large integers will
    be included in the output from `rand_number`"""
    global _pretty_numbers
    _pretty_numbers = enabled


def rand_number() -> NumberType:
    global _pretty_numbers
    result: NumberType
    if _pretty_numbers:
        min_value = 1
        max_value = 12
    else:
        min_value = -10000
        max_value = 10000
    # Use an integer?
    if _pretty_numbers or rand_bool(66):
        if _pretty_numbers or rand_bool(50):
            result = random.randint(min_value, max_value)
        else:
            result = random.randint(1, max_const)
    else:
        if rand_bool(10):
            result = random.random() * max_value
        elif rand_bool(10):
            result = random.random() * min_value
        else:
            result = random.random() * max_const

        result = truncate(result, max_decimals=1)
    return result


def truncate(value: NumberType, max_decimals: int = 3) -> NumberType:
    return float(f"%.{max_decimals}f" % (float(value)))


def rand_op() -> str:
    return operators[random.randint(0, len(operators) - 1)]


def get_rand_vars(
    num_vars: int,
    exclude_vars: Optional[List[str]] = None,
    common_variables: bool = False,
) -> List[str]:
    """Get a list of random variables, excluding the given list of hold-out variables"""
    if exclude_vars is None:
        exclude_vars = []
    if num_vars > 25:
        raise ValueError("out of range: there are only twenty-six variables")
    rand_vars: Set[str] = set()
    iters = 0
    while len(rand_vars) < num_vars:
        _rand = rand_var(common_variables)
        if _rand not in exclude_vars:
            rand_vars.add(_rand)
        iters += 1
        if iters > num_vars * 10:
            raise ValueError(
                f"Unable to fulfill request for {num_vars} random variables"
            )
    out = list(rand_vars)
    random.shuffle(out)
    return out


def gen_binomial_times_binomial(
    *,
    op: str = "+",
    min_vars: int = 1,
    max_vars: int = 2,
    simple_variables: bool = True,
    powers_probability: float = 0.33,
    like_variables_probability: float = 1.0,
) -> Tuple[str, int]:
    """Generate a binomial multiplied by another binomial.

    # Example

    ```
    (2e + 12p)(16 + 7e)
    ```

    `mathy:(2e + 12p)(16 + 7e)`
    """
    power_prob_percent = powers_probability * 100
    powers = rand_bool(power_prob_percent)
    like_vars = rand_bool(like_variables_probability * 100)

    num_terms: int = 4

    num_vars: int = random.randint(min_vars, max_vars)
    terms = [""] * 4

    # Build variables (with optional exponents)
    if like_vars is False:
        for i, var in enumerate(get_rand_vars(num_vars)):
            if powers is not False:
                terms[i] = f"{var}{maybe_power(power_prob_percent * 2)}"
            else:
                terms[i] = var
    else:
        var = rand_var()
        if powers is not False:
            var = f"{var}{maybe_power(power_prob_percent * 2)}"
        for i in range(num_vars):
            terms[i] = var
    # random.shuffle(terms)

    # Conditionally attach coefficients to each term
    for i in range(4):
        if simple_variables is True and terms[i] != "":
            continue
        terms[i] = f"{rand_number()}{terms[i]}"

    first = [terms[0], terms[2]]
    second = [terms[1], terms[3]]
    random.shuffle(first)
    random.shuffle(second)
    return f"({first[0]} + {first[1]})({second[0]} + {second[1]})", num_terms + 2


def gen_binomial_times_monomial(
    *,
    op: str = "+",
    min_vars: int = 1,
    max_vars: int = 2,
    simple_variables: bool = True,
    powers_probability: float = 0.33,
    like_variables_probability: float = 1.0,
) -> Tuple[str, int]:
    """Generate a binomial multiplied by a monomial.

    # Example

    ```
    (4x^3 + y) * 2x
    ```

    `mathy:(4x^3 + y) * 2x`
    """
    power_prob_percent = powers_probability * 100
    powers = rand_bool(power_prob_percent)
    like_vars = rand_bool(like_variables_probability * 100)

    num_terms: int = 3

    num_vars: int = random.randint(min_vars, max_vars)
    terms = [""] * num_terms

    # Build variables (with optional exponents)
    if like_vars is False:
        for i, var in enumerate(get_rand_vars(num_vars)):
            if powers is not False:
                terms[i] = f"{var}{maybe_power(power_prob_percent * 2)}"
            else:
                terms[i] = var
    else:
        var = rand_var()
        if powers is not False:
            var = f"{var}{maybe_power(power_prob_percent * 2)}"
        for i in range(num_vars):
            terms[i] = var
    # random.shuffle(terms)

    # Conditionally attach coefficients to each term
    for i in range(num_terms):
        if simple_variables is True and terms[i] != "":
            continue
        terms[i] = f"{rand_number()}{terms[i]}"

    first = [terms[0], terms[2]]
    second = terms[1]
    random.shuffle(first)
    return f"({first[0]} + {first[1]}) * {second}", num_terms


def gen_simplify_multiple_terms(
    num_terms: int,
    optional_var: bool = False,
    op: Union[List[str], str] = None,
    common_variables: bool = True,
    inner_terms_scaling: float = 0.3,
    powers_probability: float = 0.33,
    optional_var_probability: float = 0.8,
    noise_probability: float = 0.8,
    shuffle_probability: float = 0.66,
    share_var_probability: float = 0.5,
    grouping_noise_probability: float = 0.66,
    noise_terms: int = None,
) -> Tuple[str, int]:
    """Generate a polynomial problem with like terms that need to be combined and
    simplified.

    # Example

    ```
    2a + 3j - 7b + 17.2a + j
    ```

    `mathy:2a + 3j - 7b + 17.2a + j`
    """
    power_prob_percent = powers_probability * 100
    use_grouping_noise = rand_bool(grouping_noise_probability * 100) is True
    num_like_terms = max(2, int(num_terms * inner_terms_scaling))
    use_noise = rand_bool(noise_probability * 100) is True
    if num_terms <= 1:
        raise ValueError("num_terms must be at least 2")
    if num_terms == 2:
        num_like_terms = 1
    like_term_vars = get_rand_vars(num_like_terms)
    term_templates = like_term_vars[:]

    # Use shared variables with different expoonents (4x, 2x^3)
    share_var: bool = rand_bool(share_var_probability * 100) is True
    shared_var: Optional[str] = None
    adorn_offset: int = 0
    if share_var:
        # Like terms with a shared variable.
        if num_like_terms > 1 and not use_noise:
            adorn_offset = 2
            term_templates[1] = f"{term_templates[0]}{maybe_power(100)}"
        else:
            # Unlike terms with a shared variable
            shared_var = f"{term_templates[0]}{maybe_power(100)}"

    # Maybe add exponents to non-shared terms
    for i, var in enumerate(term_templates[adorn_offset:]):
        term_templates[i + adorn_offset] = f"{var}{maybe_power(power_prob_percent)}"

    complexity = num_terms

    # Repeat enough times to satisfy max_terms and trim
    term_templates *= num_terms
    term_templates = term_templates[0:num_terms]

    # Do this after repeating the term_templates (else shared would always have
    # "like" terms)
    if shared_var is not None:
        term_templates.append(shared_var)

    # sometimes add noise terms to the ends
    if use_noise:
        num_noise_terms = min(5, max(1, num_terms // 3))
        if noise_terms is not None:
            num_noise_terms = noise_terms
        noise_vars = get_rand_vars(num_noise_terms, like_term_vars)

        # When there's noise add complexity
        complexity += 1

        # We take the larger value for the left side to push the terms
        # that have to be matched to the right side of expression. This is
        # so that the model cannot use its existing knowledge about distributive
        # factoring on smaller problems to solve this problem.
        lr_terms: Tuple[int, int] = split_in_two_random(num_noise_terms)
        for i in range(lr_terms[0]):
            current = noise_vars.pop()
            term_templates.insert(0, f"{current}{maybe_power(power_prob_percent)}")
        for i in range(lr_terms[1]):
            current = noise_vars.pop()
            term_templates.append(f"{current}{maybe_power(power_prob_percent)}")

    # sometimes shuffle the terms
    if rand_bool(shuffle_probability * 100) is True:
        random.shuffle(term_templates)

    def get_op() -> str:
        if op is None:
            return rand_op()
        if isinstance(op, list):
            return random.choice(op)
        return op

    group_start = group_end = -1
    if use_grouping_noise:
        half_point = max(len(term_templates) // 2, 1)
        group_start = random.randint(0, half_point - 1)
        group_end = random.randint(half_point, len(term_templates) - 1)
        assert group_start < group_end
        assert group_start != group_end
    root_term = term_templates.pop(0)
    result = f"{maybe_number()}{root_term}"
    if use_grouping_noise and group_start == 0:
        result = f"({result}"
    for i, other_var in enumerate(term_templates):
        if optional_var and rand_bool(optional_var_probability * 100) is False:
            other_var = ""
        coefficient = rand_number() if other_var == "" else maybe_number()
        term = f"{coefficient}{other_var}"
        if use_grouping_noise:
            if i + 1 == group_start:
                term = f"({term}"
            elif i + 1 == group_end:
                term = f"{term})"
        result += f" {get_op()} {term}"
    return result, complexity


def split_in_two_random(value: int) -> Tuple[int, int]:
    """Split a given number into two smaller numbers that sum to it.
    Returns: a tuple of (lower, higher) numbers that sum to the input
    """
    factor = random.uniform(0, 1)
    left = int(factor * value)
    right = value - left
    lower = min(left, right)
    higher = max(left, right)
    assert lower + higher == value, "did not split evenly in two"
    return lower, higher


def gen_combine_terms_in_place(
    min_terms: int = 16, max_terms: int = 26, easy: bool = True, powers: bool = False
) -> Tuple[str, int]:
    """Generate a problem that puts one pair of like terms next to each other
    somewhere inside a large tree of unlike terms.

    The problem is intended to be solved in a very small number of moves, making
    training across many episodes relatively quick, and reducing the combinatorial
    explosion of branches that need to be searched to solve the task.

    The hope is that by focusing the agent on selecting the right moves inside of a
    ridiculously large expression it will learn to select actions to combine like terms
    invariant of the sequence length.

    # Example

    ```
    4y + 12j + 73q + 19k + 13z + 56l + (24x + 12x) + 43n + 17j
    ```

    `mathy:4y + 12j + 73q + 19k + 13z + 56l + (24x + 12x) + 43n + 17j`

    """

    total_terms = random.randint(min_terms, max_terms)
    var = rand_var()
    power_chance = 80 if powers is True else 0
    power = maybe_power(power_chance)
    focus_chunk = f"{maybe_number()}{var}{power} + {maybe_number()}{var}{power}"
    if easy:
        focus_chunk = f"({focus_chunk})"
    num_noise_terms = total_terms - 2
    noise_vars = get_rand_vars(num_noise_terms, [var])

    out_terms = []

    # We take the larger value for the left side to push the terms
    # that have to be matched to the right side of expression. This is
    # so that the model cannot use its existing knowledge about distributive
    # factoring on smaller problems to solve this problem.
    right_num, left_num = split_in_two_random(num_noise_terms)
    for i in range(left_num):
        current = noise_vars.pop()
        out_terms.append(f"{maybe_number()}{current}{maybe_power(power_chance)}")
    out_terms.append(focus_chunk)
    for i in range(right_num):
        current = noise_vars.pop()
        out_terms.append(f"{maybe_number()}{current}{maybe_power(power_chance)}")

    complexity = total_terms
    problem = " + ".join(out_terms)
    return problem, complexity


def gen_commute_haystack(
    min_terms: int = 5,
    max_terms: int = 8,
    commute_blockers: int = 1,
    easy: bool = True,
    powers: bool = False,
) -> Tuple[str, int]:
    """A problem with a bunch of terms that have no matches, and a single
    set of two terms that do match, but are separated by one other term.
    The challenge is to commute the terms to each other in one move.

    # Example

    ```
    4y + 12j + 73q + 19k + 13z + 24x + 56l + 12x  + 43n + 17j"
                                  ^-----------^
    ```

    `mathy:4y + 12j + 73q + 19k + 13z + 24x + 56l + 12x  + 43n + 17j`
    """
    total_terms = random.randint(min_terms, max_terms)
    num_noise_terms = max(total_terms - 2, commute_blockers)
    var = rand_var()
    noise_vars = get_rand_vars(num_noise_terms, [var])
    power_chance = 80 if powers is True else 0
    power = maybe_power(power_chance)

    # Build up the blockers to put between the like terms
    blockers = []
    for i in range(commute_blockers):
        current = noise_vars.pop()
        blockers.append(f"{maybe_number()}{current}{maybe_power(power_chance)}")

    blocks = " + ".join(blockers)
    focus_chunk = (
        f"{maybe_number()}{var}{power} + {blocks} + {maybe_number()}{var}{power}"
    )
    # About half of the time focus the agent by grouping the subtree for them
    if rand_bool(50 if easy else 10):
        focus_chunk = f"({focus_chunk})"

    out_terms = []

    # We take the larger value for the left side to push the terms
    # that have to be matched to the right side of expression. This is
    # so that the model cannot use its existing knowledge about distributive
    # factoring on smaller problems to solve this problem.
    right_num, left_num = split_in_two_random(num_noise_terms - commute_blockers)
    for i in range(left_num):
        current = noise_vars.pop()
        out_terms.append(f"{maybe_number()}{current}{maybe_power(power_chance)}")
    out_terms.append(focus_chunk)
    for i in range(right_num):
        current = noise_vars.pop()
        out_terms.append(f"{maybe_number()}{current}{maybe_power(power_chance)}")

    complexity = len(out_terms)
    problem = " + ".join(out_terms)
    return problem, complexity


def get_blocker(num_blockers: int = 1, exclude_vars: Optional[List[str]] = None) -> str:
    """Get a string of terms to place between target simplification terms
    in order to challenge the agent's ability to use commutative/associative
    rules to move terms around."""
    if exclude_vars is None:
        exclude_vars = []
    vars = get_rand_vars(num_blockers, exclude_vars)
    out_terms = []
    for i in range(num_blockers):
        out_terms.append("{}{}".format(maybe_number(), vars[i]))
    return " + ".join(out_terms)


def gen_move_around_blockers_one(
    number_blockers: int, powers_probability: float = 0.5
) -> Tuple[str, int]:
    """Two like terms separated by (n) blocker terms.

    # Example

    ```
    4x + (y + f) + x
    ```

    `mathy:4x + (y + f) + x`"""
    var = rand_var()
    power_chance = powers_probability * 100
    exp = maybe_power(power_chance)
    complexity = 2 + number_blockers
    blockers = get_blocker(number_blockers, [var])
    problem = "{}{}{} + {} + {}{}{}".format(
        maybe_number(), var, exp, blockers, maybe_number(), var, exp
    )
    return problem, complexity


def gen_move_around_blockers_two(
    number_blockers: int, powers_probability: float = 0.5
) -> Tuple[str, int]:
    """Two like terms with three blockers.

    # Example

    ```
    7a + 4x + (2f + j) + x + 3d
    ```

    `mathy:7a + 4x + (2f + j) + x + 3d`"""
    rand_vars = get_rand_vars(3)
    [one_var, two_var, three_var] = rand_vars
    complexity = 4 + number_blockers
    power_chance = powers_probability * 100
    one_exp = maybe_power(power_chance)
    two_exp = maybe_power(power_chance)
    three_exp = maybe_power(power_chance)
    problem = "{}{}{} + {}{}{} + {} + {}{}{} + {}{}{}".format(
        maybe_number(),
        one_var,
        one_exp,
        maybe_number(),
        two_var,
        two_exp,
        get_blocker(number_blockers, rand_vars),
        maybe_number(),
        two_var,
        two_exp,
        maybe_number(),
        three_var,
        three_exp,
    )
    return problem, complexity



from typing import Dict, List, Optional

from .expressions import (
    AddExpression,
    ConstantExpression,
    DivideExpression,
    EqualExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
    PowerExpression,
    SubtractExpression,
    VariableExpression,
)
from .tokenizer import (
    Token,
    TokenCloseParen,
    TokenConstant,
    TokenDivide,
    TokenEOF,
    TokenEqual,
    TokenExponent,
    TokenFunction,
    Tokenizer,
    TokenMinus,
    TokenMultiply,
    TokenNone,
    TokenOpenParen,
    TokenPlus,
    TokenVariable,
    coerce_to_number,
)
from .types import NumberType


class ParserException(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message


class InvalidExpression(ParserException):
    pass


class OutOfTokens(ParserException):
    pass


class InvalidSyntax(ParserException):
    pass


class UnexpectedBehavior(ParserException):
    pass


class TrailingTokens(ParserException):
    pass


class TokenSet:
    """TokenSet objects are bitmask combinations for checking to see
    if a token is part of a valid set."""

    tokens: int

    def __init__(self, source: int):
        self.tokens = source

    def add(self, addTokens: int) -> "TokenSet":
        """Add tokens to self set and return a TokenSet representing
        their combination of flags.  Value can be an integer or an instance
        of `TokenSet`"""
        return TokenSet(self.tokens | addTokens)

    def contains(self, type: int) -> bool:
        """Returns true if the given type is part of this set"""
        return (self.tokens & type) != 0


_FIRST_FUNCTION: TokenSet = TokenSet(TokenFunction)
_FIRST_FACTOR: TokenSet = _FIRST_FUNCTION.add(TokenVariable | TokenOpenParen)
_FIRST_FACTOR_PREFIX: TokenSet = _FIRST_FACTOR.add(TokenConstant)
_FIRST_UNARY: TokenSet = _FIRST_FACTOR_PREFIX.add(TokenMinus)
_FIRST_EXP: TokenSet = _FIRST_UNARY
_FIRST_MULT: TokenSet = _FIRST_UNARY
_FIRST_ADD: TokenSet = _FIRST_UNARY
_FIRST_EQUAL: TokenSet = _FIRST_UNARY

# Precedence checks
_IS_ADD: TokenSet = TokenSet(TokenPlus | TokenMinus)
_IS_MULT: TokenSet = TokenSet(TokenMultiply | TokenDivide)
_IS_EXP: TokenSet = TokenSet(TokenExponent)
_IS_EQUAL: TokenSet = TokenSet(TokenEqual)


# NOTE: This cannot be shared between threads because it stores state in
#       self.current_token and self.tokens
class ExpressionParser:
    """Parser for converting text into binary trees. Trees encode the order of
    operations for an input, and allow evaluating it to detemrine the expression
    value.

    ### Grammar Rules

    Symbols:
    ```
    ( )    == Non-terminal
    { }*   == 0 or more occurrences
    { }+   == 1 or more occurrences
    { }?   == 0 or 1 occurrences
    [ ]    == Mandatory (1 must occur)
    |      == logical OR
    " "    == Terminal symbol (literal)
    ```

    Non-terminals defined/parsed by Tokenizer:
    ```
    (Constant) = anything that can be parsed by `float(in)`
    (Variable) = any string containing only letters (a-z and A-Z)
    ```

    Rules:
    ```
    (Function)     = [ functionName ] "(" (AddExp) ")"
    (Factor)       = { (Variable) | (Function) | "(" (AddExp) ")" }+ { { "^" }? (UnaryExp) }?
    (FactorPrefix) = [ (Constant) { (Factor) }? | (Factor) ]
    (UnaryExp)     = { "-" }? (FactorPrefix)
    (ExpExp)       = (UnaryExp) { { "^" }? (UnaryExp) }?
    (MultExp)      = (ExpExp) { { "*" | "/" }? (ExpExp) }*
    (AddExp)       = (MultExp) { { "+" | "-" }? (MultExp) }*
    (EqualExp)     = (AddExp) { { "=" }? (AddExp) }*
    (start)        = (EqualExp)
    ```
    """  # noqa

    _parse_cache: Dict[str, MathExpression]
    _tokens_cache: Dict[str, List[Token]]
    _all_tokens: Optional[List[Token]]

    # Initialize the tokenizer.
    def __init__(self) -> None:
        self.tokenizer = Tokenizer()
        self.clear_cache()

    def clear_cache(self) -> None:
        self._tokens_cache = {}
        self._parse_cache = {}

    def tokenize(self, input_text: str) -> List[Token]:
        if input_text not in self._tokens_cache:
            self._tokens_cache[input_text] = self.tokenizer.tokenize(input_text)
        return self._tokens_cache[input_text][:]

    def parse(self, input_text: str) -> MathExpression:
        """Parse a string representation of an expression into a tree
        that can be later evaluated.

        Returns : The evaluatable expression tree.
        """
        if input_text in self._parse_cache:
            return self._parse_cache[input_text]
        self._parse_cache[input_text] = self._parse(self.tokenize(input_text))
        return self._parse_cache[input_text]

    def _parse(self, tokens: List[Token]) -> MathExpression:
        """Parse a given list of tokens into an expression tree"""
        self.tokens = tokens
        self._all_tokens = tokens[:]
        self.current_token = Token("", TokenNone)
        if not self.next():
            raise InvalidExpression("Cannot parse an empty function")

        expression: MathExpression = self.parse_equal()
        leftover = ""
        while self.current_token.type != TokenEOF:
            leftover = f"{leftover}{self.current_token.value}"
            self.next()

        if leftover != "":
            raise TrailingTokens("Trailing characters: {}".format(leftover))
        self._all_tokens = None
        return expression

    def parse_equal(self) -> MathExpression:
        if not self.check(_FIRST_ADD):
            raise InvalidSyntax("Invalid expression")

        exp = self.parse_add()
        while self.check(_IS_EQUAL):
            opType = self.current_token.type
            opValue = self.current_token.value
            self.eat(opType)
            expected = self.check(_FIRST_ADD)
            right = None
            if expected:
                right = self.parse_add()

            if not expected or not right:
                raise UnexpectedBehavior(
                    "Expected an expression after = operator, got: {}".format(
                        self.current_token.value
                    )
                )

            if opType != TokenEqual:
                raise UnexpectedBehavior(
                    "Expected plus or minus, got: {}".format(opValue)
                )

            exp = EqualExpression(exp, right)

        return exp

    def parse_add(self) -> MathExpression:
        if not self.check(_FIRST_MULT):
            raise InvalidSyntax("Invalid expression")

        exp = self.parse_mult()
        while self.check(_IS_ADD):
            opType = self.current_token.type
            opValue = self.current_token.value
            self.eat(opType)
            expected = self.check(_FIRST_MULT)
            right = None
            if expected:
                right = self.parse_mult()

            if not expected or not right:
                raise UnexpectedBehavior(
                    "Expected an expression after + or - operator, got: {}".format(
                        self.current_token.value
                    )
                )

            if opType == TokenPlus:
                exp = AddExpression(exp, right)
            elif opType == TokenMinus:
                exp = SubtractExpression(exp, right)
            else:
                raise UnexpectedBehavior(
                    "Expected plus or minus, got: {}".format(opValue)
                )

        return exp

    def parse_mult(self) -> MathExpression:
        if not self.check(_FIRST_EXP):
            raise InvalidSyntax("Invalid expression")

        exp = self.parse_exponent()
        while self.check(_IS_MULT):
            opType = self.current_token.type
            opValue = self.current_token.value
            self.eat(opType)
            expected = self.check(_FIRST_EXP)
            right = None
            if expected:
                right = self.parse_mult()

            if not expected or right is None:
                assert self._all_tokens is not None
                input_str = "".join([str(f.value) for f in self._all_tokens])
                raise InvalidSyntax(
                    f"Expected an expression after * or / operator, got: {opValue}"
                    f"\nFull input: {input_str}"
                )

            if opType == TokenMultiply:
                exp = MultiplyExpression(exp, right)
            elif opType == TokenDivide:
                exp = DivideExpression(exp, right)
            else:
                raise UnexpectedBehavior(
                    "Expected mult or divide, got: {}".format(opValue)
                )
        return exp

    def parse_exponent(self) -> MathExpression:
        if not self.check(_FIRST_UNARY):
            raise InvalidSyntax("Invalid expression")

        exp = self.parse_unary()
        if self.check(TokenSet(TokenExponent)):
            opType = self.current_token.type
            self.eat(opType)
            if not self.check(_FIRST_UNARY):
                raise InvalidSyntax("Expected an expression after ^ operator")

            right = self.parse_unary()
            if opType == TokenExponent:
                exp = PowerExpression(exp, right)
            else:
                raise UnexpectedBehavior("Expected exponent, got: {}".format(opType))
        return exp

    def parse_unary(self) -> MathExpression:
        value: NumberType = 0
        negate = False
        if self.current_token.type == TokenMinus:
            self.eat(TokenMinus)
            negate = True
        expected = self.check(_FIRST_FACTOR_PREFIX)
        exp: Optional[MathExpression] = None
        if expected:
            if self.current_token.type == TokenConstant:
                if isinstance(self.current_token.value, str):
                    value = coerce_to_number(self.current_token.value)
                else:
                    value = self.current_token.value
                # Flip parse as float/int based on whether the value text
                if negate:
                    value = -value
                    negate = False

                exp = ConstantExpression(value)
                self.eat(TokenConstant)

            if self.check(_FIRST_FACTOR):
                if exp is None:
                    exp = self.parse_factors()
                else:
                    exp = MultiplyExpression(exp, self.parse_factors())

        if not expected or exp is None:
            raise InvalidSyntax(
                "Expected a function/variable/parenthesis after - or + \n"
                f"Received : {self.current_token.value}"
            )
        if negate:
            return NegateExpression(exp)

        return exp

    def parse_factors(self) -> MathExpression:
        right = None
        found = True
        factors: List[MathExpression] = []
        while found:
            right = None
            opType = self.current_token.type
            if opType == TokenVariable:
                factors.append(VariableExpression(str(self.current_token.value)))
                self.eat(TokenVariable)
            elif opType == TokenFunction:
                factors.append(self.parse_function())
            elif opType == TokenOpenParen:
                self.eat(TokenOpenParen)
                factors.append(self.parse_add())
                self.eat(TokenCloseParen)
            else:
                raise UnexpectedBehavior(
                    "Unexpected token in Factor: {}".format(self.current_token.value)
                )

            found = self.check(_FIRST_FACTOR)

        if len(factors) == 0:
            raise InvalidExpression("No factors")

        exp: Optional[MathExpression] = None
        if self.check(_IS_EXP):
            opType = self.current_token.type
            self.eat(opType)
            if not self.check(_FIRST_UNARY):
                raise InvalidSyntax("Expected an expression after ^ operator")

            right = self.parse_unary()
            exp = PowerExpression(factors[-1], right)

        if len(factors) == 1:
            return exp or factors[0]

        while len(factors) > 0:
            if exp is None:
                exp = factors.pop(0)

            exp = MultiplyExpression(exp, factors.pop(0))

        assert exp is not None
        return exp

    def parse_function(self) -> MathExpression:
        opFn = str(self.current_token.value)
        self.eat(self.current_token.type)
        self.eat(TokenOpenParen)
        exp = self.parse_add()
        self.eat(TokenCloseParen)
        func = self.tokenizer.functions[opFn]
        if func is None:
            raise UnexpectedBehavior("Unknown Function type: {}".format(opFn))

        return func(exp)

    def next(self) -> bool:
        """Assign the next token in the queue to `self.current_token`.

        Return True if there are still more tokens in the queue, or False if there
        are no more tokens to look at."""

        if self.current_token.type == TokenEOF:
            raise OutOfTokens("Parsed beyond the end of the expression")

        self.current_token = self.tokens.pop(0)
        return self.current_token.type != TokenEOF

    def eat(self, type: int) -> bool:
        """Assign the next token in the queue to current_token if its type
        matches that of the specified parameter. If the type does not match,
        raise a syntax exception.

        Args:
            - `type` The type that your syntax expects @current_token to be
        """
        if self.current_token.type != type:
            raise InvalidSyntax("Missing: {}".format(type))

        return self.next()

    def check(self, tokens: TokenSet) -> bool:
        """Check if the `self.current_token` is a member of a set Token types

        Args:
            - `tokens` The set of Token types to check against

        `Returns` True if the `current_token`'s type is in the set else False"""

        return tokens.contains(self.current_token.type)



from typing import Optional

from ..expressions import (
    AddExpression,
    ConstantExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
    PowerExpression,
    SubtractExpression,
    VariableExpression,
)
from ..rule import BaseRule, ExpressionChangeRule


class RestateSubtractionRule(BaseRule):
    """Convert subtract operators to plus negative to allow commuting"""

    OP_SUBTRACTION = "subtraction"
    OP_ADD_NEG_CONST = "add_neg_const"
    OP_ADD_NEG_CONST_VAR = "add_neg_const_var"
    OP_ADD_NEG_CONST_VAR_EXP = "add_neg_const_var_exp"

    @property
    def name(self) -> str:
        return "Restate Subtraction"

    @property
    def code(self) -> str:
        return "RS"

    def get_type(self, node: MathExpression) -> Optional[str]:
        """Determine the configuration of the tree for this transformation.

        Support two types of tree configurations:
         - Subtraction is a subtract to be restate as a plus negation
         - PlusNegative is a plus negative const to be restated as subtraction
        """
        is_sub = isinstance(node, SubtractExpression)
        is_parent_add = isinstance(node.parent, AddExpression)
        # 4x - 2x
        if is_sub and (node.parent is None or is_parent_add):
            return RestateSubtractionRule.OP_SUBTRACTION
        # The other cases are addition and negative/negation
        is_add = isinstance(node, AddExpression)
        if not is_add:
            return None

        # + -2 = Const
        if node.right and isinstance(node.right, ConstantExpression):
            assert node.right.value is not None
            if node.right.value < 0.0:
                return RestateSubtractionRule.OP_ADD_NEG_CONST
            return None

        # + -2x = Mult(Const, Var)
        if (
            node.right
            and isinstance(node.right, MultiplyExpression)
            and isinstance(node.right.left, ConstantExpression)
            and isinstance(node.right.right, VariableExpression)
        ):
            assert node.right.left.value is not None
            if node.right.left.value < 0.0:
                return RestateSubtractionRule.OP_ADD_NEG_CONST_VAR
            return None

        # + -2x^3 = Mult(Const, Exp(Var, Const))
        if (
            isinstance(node.right, MultiplyExpression)
            and isinstance(node.right.left, ConstantExpression)
            and isinstance(node.right.right, PowerExpression)
        ):
            assert node.right.left.value is not None
            if node.right.left.value < 0.0:
                return RestateSubtractionRule.OP_ADD_NEG_CONST_VAR_EXP
            return None

        return None

    def can_apply_to(self, node: MathExpression) -> bool:
        tree_type = self.get_type(node)
        return tree_type is not None

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node)
        tree_type = self.get_type(node)
        assert tree_type is not None, "call can_apply_to before applying a rule"
        change.save_parent()  # connect result to node.parent
        result: MathExpression

        # Simple case of subtract to plus negate
        if tree_type == RestateSubtractionRule.OP_SUBTRACTION:
            result = AddExpression(node.left, NegateExpression(node.right))
        else:
            # Plus negative to subtract cases
            assert node.left is not None and node.right is not None
            result = SubtractExpression(node.left, node.right.clone())
            if tree_type == RestateSubtractionRule.OP_ADD_NEG_CONST:
                assert (
                    result.right is not None
                    and isinstance(result.right, ConstantExpression)
                    and result.right.value is not None
                )
                result.right.value = -result.right.value
            elif tree_type == RestateSubtractionRule.OP_ADD_NEG_CONST_VAR:
                assert (
                    result.right is not None
                    and isinstance(result.right.left, ConstantExpression)
                    and result.right.left.value is not None
                )
                result.right.left.value = -result.right.left.value
            elif tree_type == RestateSubtractionRule.OP_ADD_NEG_CONST_VAR_EXP:
                assert (
                    result.right is not None
                    and isinstance(result.right.left, ConstantExpression)
                    and result.right.left.value is not None
                )
                result.right.left.value = -result.right.left.value
        result.set_changed()  # mark this node as changed for visualization
        return change.done(result)



from typing import Any, Callable, List, Optional, TypeVar, Union

from .types import Literal

NodeType = TypeVar("NodeType", bound="BinaryTreeNode")
try:
    NodeType.__doc__ = "Template type that inherits from BinaryTreeNode."
except AttributeError:
    ...  # Boo!

# ## Constants

# Return this from a node visit function to abort a tree visit.
STOP: Literal["stop"] = "stop"
# The constant representing the left child side of a node.
LEFT: Literal["left"] = "left"
# The constant representing the right child side of a node.
RIGHT: Literal["right"] = "right"


SideType = Union[Literal["left"], Literal["right"]]
VisitStop = Literal["stop"]
VisitDataType = TypeVar("VisitDataType", bound=Any)
try:
    VisitDataType.__doc__ = "Template type of user data passed to visit functions."
except AttributeError:
    ...  # Boo!
VisitFunction = Callable[[NodeType, int, VisitDataType], Optional[VisitStop]]


class BinaryTreeNode:
    """
    The binary tree node is the base node for all of our trees, and provides a
    rich set of methods for constructing, inspecting, and modifying them.
    The node itself defines the structure of the binary tree, having left and right
    children, and a parent.
    """

    _idCounter = 0

    # Tree layout mutations. Thanks, 2009 Justin. :( :(
    x: Optional[float]
    y: Optional[float]
    offset: Optional[float]
    level: Optional[int]
    thread: Optional["BinaryTreeNode"]

    left: Optional["BinaryTreeNode"]
    right: Optional["BinaryTreeNode"]
    parent: Optional["BinaryTreeNode"]

    #  Allow specifying children in the constructor
    def __init__(
        self,
        left: "BinaryTreeNode" = None,
        right: "BinaryTreeNode" = None,
        parent: "BinaryTreeNode" = None,
        id: Optional[str] = None,
    ):
        if id is None:
            BinaryTreeNode._idCounter = BinaryTreeNode._idCounter + 1
            id = f"mn-{BinaryTreeNode._idCounter}"
        self.id = id
        self.left = None
        self.right = None
        self.set_left(left)
        self.set_right(right)
        self.parent = parent

    def clone(self) -> NodeType:
        """Create a clone of this tree"""
        result: NodeType = self.__class__()  # type:ignore
        result.id = self.id
        if self.left:
            result.set_left(self.left.clone())  # type:ignore

        if self.right:
            result.set_right(self.right.clone())  # type:ignore
        return result

    def is_leaf(self) -> bool:
        """Is this node a leaf?  A node is a leaf if it has no children."""
        return not self.left and not self.right

    def __str__(self) -> str:
        """Serialize the node as a str"""
        return "{} {}".format(self.left, self.right)

    @property
    def name(self) -> str:
        """Human readable name for this node."""
        return "BinaryTreeNode"

    def rotate(self: NodeType) -> NodeType:
        """
        Rotate a node, changing the structure of the tree, without modifying
        the order of the nodes in the tree.
        """
        node = self
        parent = self.parent
        if not node or not parent:
            return self

        grand_parent = parent.parent
        if node == parent.left:
            parent.set_left(node.right)
            node.right = parent
            parent.parent = node
        else:
            parent.set_right(node.left)
            node.left = parent
            parent.parent = node

        node.parent = grand_parent
        if not grand_parent:
            return self

        if parent == grand_parent.left:
            grand_parent.left = node
        else:
            grand_parent.right = node
        return self

    def visit_preorder(
        self: NodeType,
        visit_fn: VisitFunction,
        depth: int = 0,
        data: Optional[VisitDataType] = None,
    ) -> Optional[VisitStop]:
        """Visit the tree preorder, which visits the current node, then its left
        child, and then its right child.

        *Visit -> Left -> Right*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP

        if self.left and self.left.visit_preorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if self.right and self.right.visit_preorder(visit_fn, depth + 1, data) == STOP:
            return STOP
        return None

    def visit_inorder(
        self: NodeType,
        visit_fn: VisitFunction,
        depth: int = 0,
        data: Optional[VisitDataType] = None,
    ) -> Optional[VisitStop]:
        """Visit the tree inorder, which visits the left child, then the current node,
        and then its right child.

        *Left -> Visit -> Right*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if self.left and self.left.visit_inorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP

        if self.right and self.right.visit_inorder(visit_fn, depth + 1, data) == STOP:
            return STOP
        return None

    def visit_postorder(
        self: NodeType,
        visit_fn: VisitFunction,
        depth: int = 0,
        data: Optional[VisitDataType] = None,
    ) -> Optional[VisitStop]:
        """Visit the tree postorder, which visits its left child, then its right child,
        and finally the current node.

        *Left -> Right -> Visit*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if self.left and self.left.visit_postorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if self.right and self.right.visit_postorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP
        return None

    def get_root(self: NodeType) -> NodeType:
        """Return the root element of this tree"""
        result = self
        while result.parent:
            result = result.parent  # type:ignore

        return result

    # **Child Management**
    #
    # Methods for setting the children on this expression.  These take care of
    # making sure that the proper parent assignments also take place.

    def set_left(
        self: NodeType,
        child: Optional["BinaryTreeNode"] = None,
        clear_old_child_parent: bool = False,
    ) -> NodeType:
        """Set the left node to the passed `child`"""
        if child == self:
            raise ValueError("nodes cannot be their own children")
        if self.left is not None and clear_old_child_parent:
            self.left.parent = None
        self.left = child
        if self.left:
            self.left.parent = self

        return self

    def set_right(
        self: NodeType,
        child: Optional["BinaryTreeNode"] = None,
        clear_old_child_parent: bool = False,
    ) -> NodeType:
        """Set the right node to the passed `child`"""
        if child == self:
            raise ValueError("nodes cannot be their own children")
        if self.right is not None and clear_old_child_parent:
            self.right.parent = None
        self.right = child
        if self.right:
            self.right.parent = self

        return self

    def get_side(self, child: NodeType) -> SideType:
        """Determine whether the given `child` is the left or right child of this
        node"""
        if child == self.left:
            return LEFT

        if child == self.right:
            return RIGHT

        raise ValueError("BinaryTreeNode.get_side: not a child of this node")

    def set_side(self, child: NodeType, side: SideType) -> NodeType:
        """Set a new `child` on the given `side`"""
        if side == LEFT:
            return self.set_left(child)  # type:ignore

        if side == RIGHT:
            return self.set_right(child)  # type:ignore

        raise ValueError("BinaryTreeNode.set_side: Invalid side")

    def get_children(self) -> List[NodeType]:
        """Get children as an array.  If there are two children, the first object will
        always represent the left child, and the second will represent the right."""
        result: List[Any] = []
        if self.left:
            result.append(self.left)

        if self.right:
            result.append(self.right)

        return result

    def get_sibling(self: NodeType) -> Optional[NodeType]:
        """Get the sibling node of this node.  If there is no parent, or the node
        has no sibling, the return value will be None."""
        if not self.parent:
            return None

        if self.parent.left == self:
            return self.parent.right  # type:ignore

        if self.parent.right == self:
            return self.parent.left  # type:ignore

        return None



from .associative_property import AssociativeSwapRule  # noqa
from .commutative_property import CommutativeSwapRule  # noqa
from .constants_simplify import ConstantsSimplifyRule  # noqa
from .distributive_factor_out import DistributiveFactorOutRule  # noqa
from .distributive_multiply_across import DistributiveMultiplyRule  # noqa
from .restate_subtraction import RestateSubtractionRule  # noqa
from .variable_multiply import VariableMultiplyRule  # noqa



from typing import Optional, Tuple

from ..expressions import (
    AddExpression,
    ConstantExpression,
    MathExpression,
    MultiplyExpression,
    PowerExpression,
    VariableExpression,
)
from ..rule import BaseRule, ExpressionChangeRule
from ..util import TermEx, get_term_ex

_POS_SIMPLE = "simple"
_POS_CHAINED = "chained"
_POS_CHAINED_LEFT_RIGHT = "chained_left_right"


class VariableMultiplyRule(BaseRule):
    r"""
    This restates `x^b * x^d` as `x^(b + d)` which has the effect of isolating
    the exponents attached to the variables, so they can be combined.

        1. When there are two terms with the same base being multiplied together, their
           exponents are added together. "x * x^3" = "x^4" because "x = x^1" so
           "x^1 * x^3 = x^(1 + 3) = x^4"

        TODO: 2. When there is a power raised to another power, they can be combined by
                 multiplying the exponents together. "x^(2^2) = x^4"

    The rule identifies terms with explicit and implicit powers, so the following
    transformations are all valid:

    Explicit powers: x^b * x^d = x^(b+d)

              *
             / \
            /   \          ^
           /     \    =   / \
          ^       ^      x   +
         / \     / \        / \
        x   b   x   d      b   d


    Implicit powers: x * x^d = x^(1 + d)

            *
           / \
          /   \          ^
         /     \    =   / \
        x       ^      x   +
               / \        / \
              x   d      1   d

    """

    @property
    def name(self) -> str:
        return "Variable Multiplication"

    @property
    def code(self) -> str:
        return "VM"

    def get_type(self, node: MathExpression) -> Optional[Tuple[str, TermEx, TermEx]]:
        """Determine the configuration of the tree for this transformation.

        Support two types of tree configurations:
         - Simple is where the node's left and right children are exactly
           terms that can be multiplied together.
         - Chained is where the node's left child is a term, but the right
           child is a continuation of a more complex term, as indicated by
           the presence of another Multiply node. In this case the left child
           of the next multiply node is the target.

        Structure:
         - Simple node(mult),node.left(term),node.right(term)
         - Chained node(mult),node.left(term),node.right(mult),node.right.left(term)
        """
        if not isinstance(node, MultiplyExpression):
            return None

        left_term = get_term_ex(node.left)
        right_term = get_term_ex(node.right)
        # Both children are multiplications
        if isinstance(node.left, MultiplyExpression) and isinstance(
            node.right, MultiplyExpression
        ):
            # Check '(36c^6 * u^3) * 7u^3'
            # node ----------------^
            # node.left.right = u^3
            # node.right = 7u^3
            chain_left_term = get_term_ex(node.left.right)
            if chain_left_term is not None and right_term is not None:
                if chain_left_term.variable == right_term.variable:
                    return (
                        _POS_CHAINED_LEFT_RIGHT,
                        chain_left_term,
                        right_term,
                    )

        # Left node in both cases is always resolved as a term.
        if left_term is None or left_term.variable is None:
            return None
        chained = False
        if right_term is None and isinstance(node.right, MultiplyExpression):
            chained = True
            right_term = get_term_ex(node.right.left)
        if right_term is None or right_term.variable is None:
            return None
        # Variables need to match because we're summing exponents.
        #
        # e.g. "4x * 2y * 5x * 3y = 120 * x^2 * y^2"
        if left_term.variable != right_term.variable:
            return None
        if chained is True:
            return _POS_CHAINED, left_term, right_term
        return _POS_SIMPLE, left_term, right_term

    def can_apply_to(self, node: MathExpression) -> bool:
        if not isinstance(node, MultiplyExpression):
            return False
        type_tuple = self.get_type(node)
        if type_tuple is None:
            return False
        return True

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node).save_parent()
        type_tuple = self.get_type(node)
        assert type_tuple is not None
        tree_position, left_term, right_term = type_tuple
        assert left_term is not None
        assert right_term is not None

        left_exp = 1 if left_term.exponent is None else left_term.exponent
        right_exp = 1 if right_term.exponent is None else right_term.exponent

        # exponential inside term, e.g. given input "2x^2 * 4x" is "(2 + 1)"
        exponent_sum = AddExpression(
            ConstantExpression(left_exp), ConstantExpression(right_exp)
        )
        # The whole power term, e.g. given input "2x^2 * 4x" is "x^(2 + 1)"
        power_term = PowerExpression(
            VariableExpression(left_term.variable), exponent_sum
        )
        result: MathExpression = power_term
        # If either term has a coefficient we have to include a second term
        # in the output
        has_left_coefficient = left_term.coefficient is not None
        has_right_coefficient = right_term.coefficient is not None
        coefficient_term: Optional[MathExpression] = None
        if has_left_coefficient or has_right_coefficient:
            left_const = 1 if left_term.coefficient is None else left_term.coefficient
            right_const = (
                1 if right_term.coefficient is None else right_term.coefficient
            )
            if has_left_coefficient and has_right_coefficient:
                # "(2 * 4)" both terms have coefficients
                coefficient_term = MultiplyExpression(
                    ConstantExpression(left_const), ConstantExpression(right_const)
                )
            else:
                # "2" only one term had a coefficient
                const_value = left_const if has_left_coefficient else right_const
                coefficient_term = ConstantExpression(const_value)

        # Check for chained results to keep from adding unnecessary parens
        if tree_position == _POS_CHAINED:
            assert node.right is not None
            keep_child = node.right.right
            # Start at the right and work up in the tree
            #
            # First the trailing part of the expression to the right of the
            # node we're changing.
            result = MultiplyExpression(power_term, keep_child)

            # Next, include the coefficient term (if any)
            if coefficient_term is not None:
                if isinstance(coefficient_term, MultiplyExpression):
                    result = MultiplyExpression(coefficient_term.right, result)
                    result = MultiplyExpression(coefficient_term.left, result)
                else:
                    result = MultiplyExpression(coefficient_term, result)
        elif tree_position == _POS_CHAINED_LEFT_RIGHT:
            assert (
                node.left is not None
            ), "invalid left node should have been rejected by can_apply"
            keep_child = node.left.left

            # Next, include the coefficient term (if any)
            result = power_term
            if coefficient_term is not None:
                if isinstance(coefficient_term, MultiplyExpression):
                    result = MultiplyExpression(
                        coefficient_term.right,
                        MultiplyExpression(coefficient_term.left, power_term),
                    )
                else:
                    result = MultiplyExpression(coefficient_term, power_term)

            # Start at the right and work up in the tree
            #
            # First the trailing part of the expression to the right of the
            # node we're changing.
            result = MultiplyExpression(keep_child, result)

        else:
            # "(2 * 4) * x^(2 + 1)"
            if coefficient_term is not None:
                result = MultiplyExpression(coefficient_term, power_term)
            else:
                result = power_term

        # Mark all nodes in the result as having changed
        result.all_changed()

        change.done(result)
        return change



from typing import Callable, Dict, List, Optional, Type, Union

from .expressions import FunctionExpression
from .types import NumberType

# Define the known types of tokens for the Tokenizer.
TokensMap: Dict[str, int] = {
    "None": 1 << 0,
    "Constant": 1 << 1,
    "Variable": 1 << 2,
    "Plus": 1 << 3,
    "Minus": 1 << 4,
    "Multiply": 1 << 5,
    "Divide": 1 << 6,
    "Exponent": 1 << 7,
    "Factorial": 1 << 8,
    "OpenParen": 1 << 9,
    "CloseParen": 1 << 10,
    "Function": 1 << 11,
    "Equal": 1 << 12,
    "Pad": 1 << 13,
    "EOF": 1 << 14,
}
MaxToken = len(TokensMap)

TokenNone = TokensMap["None"]
TokenConstant = TokensMap["Constant"]
TokenVariable = TokensMap["Variable"]
TokenPlus = TokensMap["Plus"]
TokenMinus = TokensMap["Minus"]
TokenMultiply = TokensMap["Multiply"]
TokenDivide = TokensMap["Divide"]
TokenExponent = TokensMap["Exponent"]
TokenFactorial = TokensMap["Factorial"]
TokenOpenParen = TokensMap["OpenParen"]
TokenCloseParen = TokensMap["CloseParen"]
TokenFunction = TokensMap["Function"]
TokenEqual = TokensMap["Equal"]
TokenPad = TokensMap["Pad"]
TokenEOF = TokensMap["EOF"]


class Token:
    value: Union[str, int, float]
    type: int

    def __init__(self, value: Union[str, int, float], type: int):
        self.value = value
        self.type = type

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f'(type={self.type}, value="{self.value}")'


class TokenContext:
    tokens: List[Token]
    index: int
    buffer: str
    chunk: str

    def __init__(
        self,
        *,
        tokens: Optional[List[Token]] = None,
        index: int = 0,
        buffer: str = "",
        chunk: str = "",
    ):
        self.tokens = tokens if tokens is not None else []
        self.index = index
        self.buffer = buffer
        self.chunk = chunk


class Tokenizer:
    """The Tokenizer produces a list of tokens from an input string."""

    exclude_padding: bool
    functions: Dict[str, Type[FunctionExpression]]

    def __init__(self, exclude_padding: bool = True):
        self.exclude_padding = exclude_padding
        self.find_functions()

    # Populate the `@functions` object with all known `FunctionExpression`s
    # in Expressions
    def find_functions(self) -> None:
        self.functions = {}
        # for (key in Expressions) {
        #   val = Expressions[key];
        #   check = {};
        #   if (check.toString.call(val) != "[object Function]") {
        #     continue;
        #   }
        #   inst = val();
        #   if (not (inst instanceof FunctionExpression)) {
        #     continue;
        #   }
        #   if (`${inst}` === "") {
        #     continue;
        #   }
        #   self.functions[`${inst}`] = val;
        # }

    # ###Token Utilities

    def is_alpha(self, c: str) -> bool:
        """Is this character a letter"""
        return ("a" <= c and c <= "z") or ("A" <= c and c <= "Z")

    def is_number(self, c: str) -> bool:
        """Is this character a number"""
        return "." == c or ("0" <= c and c <= "9")

    def eat_token(self, context: TokenContext, typeFn: Callable[[str], bool]) -> str:
        """Eat all of the tokens of a given type from the front of the stream
        until a different type is hit, and return the text."""
        res = ""
        for ch in list(context.chunk):
            if not typeFn(ch):
                return res
            res = res + str(ch)

        return res

    def tokenize(self, buffer: str) -> List[Token]:
        """Return an array of `Token`s from a given string input.
        This throws an exception if an unknown token type is found in the input."""
        context = TokenContext(buffer=buffer, chunk=str(buffer))
        while context.chunk and (
            self.identify_constants(context)
            or self.identify_alphas(context)
            or self.identify_operators(context)
        ):
            context.chunk = context.buffer[context.index :]  # noqa

        context.tokens.append(Token("", TokenEOF))
        return context.tokens

    def identify_operators(self, context: TokenContext) -> bool:
        """Identify and tokenize operators."""
        ch = context.chunk[0]
        if ch == " " or ch == "\t" or ch == "\r" or ch == "\n":
            # NOTE: originally introduced this to include padding for token prediction
            if self.exclude_padding is False:
                context.tokens.append(Token(ch, TokenPad))
        elif ch == "+":
            context.tokens.append(Token("+", TokenPlus))
        elif ch == "-" or ch == "–":
            context.tokens.append(Token("-", TokenMinus))
        elif ch == "*":
            context.tokens.append(Token("*", TokenMultiply))
        elif ch == "/":
            context.tokens.append(Token("/", TokenDivide))
        elif ch == "^":
            context.tokens.append(Token("^", TokenExponent))
        elif ch == "!":
            context.tokens.append(Token("!", TokenFactorial))
        elif ch == "(" or ch == "[":
            context.tokens.append(Token("(", TokenOpenParen))
        elif ch == ")" or ch == "]":
            context.tokens.append(Token(")", TokenCloseParen))
        elif ch == "=":
            context.tokens.append(Token("=", TokenEqual))
        else:
            raise Exception(f'Invalid token "{ch}" in expression: {context.buffer}')
        context.index = context.index + 1
        return True

    def identify_alphas(self, context: TokenContext) -> int:
        """Identify and tokenize functions and variables."""
        if not self.is_alpha(context.chunk[0]):
            return False

        variable = self.eat_token(context, self.is_alpha)
        if variable in self.functions:
            context.tokens.append(Token(variable, TokenFunction))
        else:
            # Each letter is its own variable
            for c in variable:
                context.tokens.append(Token(c, TokenVariable))

        context.index += len(variable)
        return len(variable)

    def identify_constants(self, context: TokenContext) -> int:
        """Identify and tokenize a constant number."""
        if not self.is_number(context.chunk[0]):
            return 0

        val = self.eat_token(context, self.is_number)
        context.tokens.append(Token(val, TokenConstant))
        context.index += len(val)
        return len(val)


def coerce_to_number(value: str) -> NumberType:
    return float(value) if "e" in value or "." in value else int(value)



from . import about  # noqa
from .expressions import *  # noqa
from .layout import *  # noqa
from .parser import *  # noqa
from .rule import *  # noqa
from .tokenizer import *  # noqa
from .tree import *  # noqa
from .util import *  # noqa



import pytest

from mathy_core import (
    AbsExpression,
    AddExpression,
    BinaryExpression,
    ConstantExpression,
    DivideExpression,
    EqualExpression,
    ExpressionParser,
    FunctionExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
    PowerExpression,
    SgnExpression,
    SubtractExpression,
    UnaryExpression,
    VariableExpression,
)


def test_expressions_get_children():
    constant = ConstantExpression(4)
    variable = VariableExpression("x")
    expr = AddExpression(constant, variable)
    # expect two children for add expression
    assert len(expr.get_children()) == 2
    # when both children are present, the 0 index should be the left child
    assert expr.get_children()[0] == constant
    assert expr.evaluate({"x": 10}) == 14


def test_expressions_type_id_abstract():
    expr = MathExpression()
    with pytest.raises(NotImplementedError):
        expr.type_id


def test_expressions_evaluate_abstract():
    expr = MathExpression()
    with pytest.raises(NotImplementedError):
        expr.evaluate()


def test_expressions_add_class():
    expr = VariableExpression("x")
    expr.add_class("as_string")
    expr.add_class(["many_classes", "as_list"])
    math_ml = expr.to_math_ml()
    assert "as_string" in math_ml
    assert "many_classes" in math_ml
    assert "as_list" in math_ml


def test_expressions_clear_classes():
    expr = VariableExpression("x")
    expr.add_class("as_string")
    expr.to_math_ml()
    assert "as_string" in expr.classes
    expr.clear_classes()
    assert "as_string" not in expr.classes


@pytest.mark.parametrize(
    "node_instance",
    [
        AddExpression(ConstantExpression(3), ConstantExpression(1)),
        SubtractExpression(ConstantExpression(3), ConstantExpression(3)),
        MultiplyExpression(ConstantExpression(3), ConstantExpression(3)),
        DivideExpression(ConstantExpression(3), ConstantExpression(3)),
        NegateExpression(ConstantExpression(3)),
        ConstantExpression(3),
        SgnExpression(ConstantExpression(-1)),
        SgnExpression(ConstantExpression(0)),
        SgnExpression(ConstantExpression(1)),
        AbsExpression(ConstantExpression(-1)),
        PowerExpression(VariableExpression("x"), ConstantExpression(3)),
    ],
)
def test_expressions_common_properties_and_methods(node_instance: MathExpression):
    assert node_instance.type_id is not None
    assert node_instance.name is not None
    assert str(node_instance) != ""
    node_instance.evaluate({"x": 2})


@pytest.mark.parametrize(
    "node_instance",
    [
        BinaryExpression(ConstantExpression(1), ConstantExpression(1)),
        UnaryExpression(ConstantExpression(1)),
        FunctionExpression(ConstantExpression(1)),
    ],
)
def test_expressions_abstract_properties_and_methods(node_instance: MathExpression):
    with pytest.raises(NotImplementedError):
        node_instance.evaluate()


def test_expressions_equality_evaluate_error():
    one = VariableExpression("x")
    two = ConstantExpression(2)
    expr = EqualExpression(one, two)
    with pytest.raises(ValueError):
        expr.evaluate()
    with pytest.raises(ValueError):
        expr.operate(one, two)  # type:ignore


def test_expressions_binary_errors():
    child = BinaryExpression()
    with pytest.raises(NotImplementedError):
        child.name
    with pytest.raises(ValueError):
        child.evaluate()


def test_expressions_unary_specify_child_side():
    child = ConstantExpression(1337)
    expr = UnaryExpression(child, child_on_left=False)
    assert expr.get_child() == child
    assert expr.left is None
    assert expr.right == child


def test_expressions_unary_evaluate_errors():
    ConstantExpression(1337)
    expr = UnaryExpression(None)
    with pytest.raises(ValueError):
        expr.evaluate()


@pytest.mark.parametrize(
    "text", ["4/x^3+2-7x*12=0", "abs(-4) + abs(34)", "-sgn(-1) / sgn(2)", "sgn(0)"]
)
def test_expressions_to_math_ml(text: str):
    expr = ExpressionParser().parse(text)
    ml_string = expr.to_math_ml()
    assert "<math xmlns='http:#www.w3.org/1998/Math/MathML'>" in ml_string
    assert "</math>" in ml_string


def test_expressions_find_id():
    expr: MathExpression = ExpressionParser().parse("4 / x")
    node: MathExpression = expr.find_type(VariableExpression)[0]
    assert expr.find_id(node.id) == node


@pytest.mark.parametrize("visit_order", ["preorder", "inorder", "postorder"])
def test_expressions_to_list(visit_order: str):
    expr: MathExpression = ExpressionParser().parse("4 / x")
    assert len(expr.to_list(visit_order)) == 3


def test_expressions_to_list_errors():
    expr: MathExpression = ExpressionParser().parse("4 / x")
    with pytest.raises(ValueError):
        expr.to_list("invalid")


def test_expressions_clone():
    constant = ConstantExpression(4)
    assert constant.value == 4
    assert constant.clone().value == 4


def test_expressions_clone_root():
    a = ConstantExpression(1100)
    b = ConstantExpression(100)
    sub = SubtractExpression(a, b)
    c = ConstantExpression(300)
    add = AddExpression(sub, c)
    d = ConstantExpression(37)
    expr = AddExpression(add, d)
    assert expr.evaluate() == 1337


def test_expressions_function_exceptions():
    x = FunctionExpression()
    with pytest.raises(NotImplementedError):
        x.name


def test_expressions_variable_exceptions():
    x = VariableExpression(None)
    with pytest.raises(ValueError):
        str(x)
    with pytest.raises(ValueError):
        VariableExpression("x").evaluate({})
    with pytest.raises(ValueError):
        x.to_math_ml()



from typing import Dict, List, Optional, Tuple, Type, Union, cast

import numpy as np
from colr import color

from .tree import STOP, BinaryTreeNode, NodeType, VisitDataType, VisitStop
from .types import NumberType

OOO_FUNCTION = 4
OOO_PARENS = 3
OOO_EXPONENT = 2
OOO_MULTDIV = 1
OOO_ADDSUB = 0
OOO_INVALID = -1


MathTypeKeys = {
    "empty": 0,
    "negate": 1,
    "equal": 2,
    "add": 3,
    "subtract": 4,
    "multiply": 5,
    "divide": 6,
    "power": 7,
    "term_root": 8,
    "term_connector": 9,
    "constant": 10,
    "sgn": 11,
    "abs": 12,
    # NOTE: reserved 13-25 for future expression types (such as functions)
    "variable": 26,
    "variable_a": 27,
    "variable_b": 28,
    "variable_c": 29,
    "variable_d": 30,
    "variable_e": 31,
    "variable_f": 32,
    "variable_g": 33,
    "variable_h": 34,
    "variable_i": 35,
    "variable_j": 36,
    "variable_k": 37,
    "variable_l": 38,
    "variable_m": 39,
    "variable_n": 40,
    "variable_o": 41,
    "variable_p": 42,
    "variable_q": 43,
    "variable_r": 44,
    "variable_s": 45,
    "variable_t": 46,
    "variable_u": 47,
    "variable_v": 48,
    "variable_w": 49,
    "variable_x": 50,
    "variable_y": 51,
    "variable_z": 52,
}


# The maximum value in type keys (for one-hot encoding)
MathTypeKeysMax = max(MathTypeKeys.values()) + 1


class MathExpression(BinaryTreeNode):
    """Math tree node with helpers for manipulating expressions.

    `mathy:x+y=z`
    """

    left: Optional["MathExpression"]
    right: Optional["MathExpression"]
    parent: Optional["MathExpression"]
    r_index: Optional[int]

    @property
    def raw(self) -> str:
        """raw text representation of the expression."""
        return str(self)

    @property
    def type_id(self) -> int:
        raise NotImplementedError("must be implemented in subclass")

    @property
    def terminal_text(self) -> str:
        """Text output of this node that includes terminal color codes that
        highlight which nodes have been changed in this tree as a result of
        a transformation."""

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:
            node._rendering_change = data
            return None

        self.visit_inorder(visit_fn, data=True)
        result = str(self)
        self.visit_inorder(visit_fn, data=False)
        return result

    @property
    def color(self) -> str:
        """Color to use for this node when rendering it as changed with
        `.terminal_text`"""
        return "green"

    _rendering_change: bool
    _changed: bool
    classes: List[str]
    cloned_node: Optional["MathExpression"]
    cloned_target: Optional[str]

    def __init__(
        self,
        id: str = None,
        left: "MathExpression" = None,
        right: "MathExpression" = None,
        parent: "MathExpression" = None,
    ):
        super().__init__(left, right, parent, id)
        self._rendering_change = False
        self._changed = False
        self.classes = [self.id]
        self.cloned_node = None
        self.cloned_target = ""

    def evaluate(self, context: Dict[str, NumberType] = None) -> NumberType:
        """Evaluate the expression, resolving all variables to constant values"""
        raise NotImplementedError("must be implemented in subclass")

    def set_changed(self) -> None:
        """Mark this node as having been changed by the application of a Rule"""
        self._changed = True

    def all_changed(self) -> None:
        """Mark this node and all of its children as changed"""

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:
            node.set_changed()
            return None

        self.visit_inorder(visit_fn)

    def with_color(self, text: str, style: str = "bright") -> str:
        """Render a string that is colored if something has changed"""
        if self._rendering_change is True and self._changed is True:
            return color(text, fore=self.color, style=style)
        return text

    def add_class(self, classes: Union[List[str], str]) -> "MathExpression":
        """Associate a class name with an expression. This class name will be
        attached to nodes when the expression is converted to a capable output
        format.

        See #MathExpression.to_math_ml_fragment"""
        class_list: List[str]
        if isinstance(classes, str):
            class_list = [classes]
        else:
            class_list = classes
        self.classes = list(set(self.classes).union(class_list))
        return self

    def to_list(self, visit: str = "preorder") -> List["MathExpression"]:
        """Convert this node hierarchy into a list."""
        results = []

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:

            results.append(node)
            return None

        if visit == "inorder":
            self.visit_inorder(visit_fn)
        elif visit == "preorder":
            self.visit_preorder(visit_fn)
        elif visit == "postorder":
            self.visit_postorder(visit_fn)
        else:
            raise ValueError(f"invalid visit order: {visit}")
        return results

    def clear_classes(self) -> None:
        """Clear all the classes currently set on the nodes in this expression."""

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:

            node.classes = []
            return None

        self.visit_inorder(visit_fn)

    def find_type(self, instanceType: Type[NodeType]) -> List[NodeType]:
        """Find an expression in this tree by type.

        - instanceType: The type to check for instances of

        Returns the found #MathExpression objects of the given type.
        """
        results: List[NodeType] = []

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:

            if isinstance(node, instanceType):
                results.append(node)  # type:ignore
            return None

        self.visit_inorder(visit_fn)
        return results

    def find_id(self, id: str) -> Optional["MathExpression"]:
        """Find an expression by its unique ID.

        Returns: The found #MathExpression or `None`
        """
        result: Optional[MathExpression] = None

        def visit_fn(
            node: MathExpression, depth: int, data: VisitDataType
        ) -> Optional[VisitStop]:

            nonlocal result
            if node.id == id:
                result = node
                return STOP
            return None

        self.visit_inorder(visit_fn)
        return result

    def to_math_ml_fragment(self) -> str:
        """Convert this single node into MathML."""
        return ""

    def to_math_ml(self) -> str:
        """Convert this expression into a MathML container."""
        return "\n".join(
            [
                "<math xmlns='http:#www.w3.org/1998/Math/MathML'>",
                self.to_math_ml_fragment(),
                "</math>",
            ]
        )

    def make_ml_tag(self, tag: str, content: str, classes: List[str] = []) -> str:
        """Make a MathML tag for the given content while respecting the node's given
        classes.

        # Arguments
        tag (str): The ML tag name to create.
        content (str): The ML content to place inside of the tag.
        classes (List[str]) An array of classes to attach to this tag.

        # Returns
        (str): A MathML element with the given tag, content, and classes
        """
        classes_attr = ""
        if len(classes) > 0:
            classes_attr = f' class="{" ".join(classes)}"'
        return f"<{tag}{classes_attr}>{content}</{tag}>"

    def path_to_root(self) -> str:
        """Generate a namespaced path key to from the current node to the root.
        This key can be used to identify a node inside of a tree."""
        points: List[str] = []

        def path_mark(node: MathExpression) -> None:
            points.append(node.__class__.__name__)

        node = self
        path_mark(node)
        while node.parent:
            node = node.parent
            path_mark(node)
        return ".".join(points)

    def clone_from_root(self, node: "MathExpression" = None) -> "MathExpression":
        """Clone this node including the entire parent hierarchy that it has. This
        is useful when you want to clone a subtree and still maintain the overall
        hierarchy.

        # Arguments
        node (MathExpression): The node to clone.

        # Returns
        (MathExpression): The cloned node.
        """
        node = node if node is not None else self
        self.cloned_node = None
        self.cloned_target = node.path_to_root()
        result = node.get_root().clone()
        if not self.cloned_node:  # pragma: nocover
            print("While cloning root of: {}".format(node))
            print(" Which is this       : {}".format(node.get_root()))
            print("Did not set the clone: {}".format(self.cloned_node))
            raise Exception("cloning root hierarchy did not clone this node")

        result = self.cloned_node
        self.cloned_node = None
        self.cloned_target = None
        return result

    def clone(self) -> "MathExpression":  # type:ignore[override]
        """A specialization of the clone method that can track and report a cloned
        subtree node.

        See #MathExpression.clone_from_root for more details."""
        result: MathExpression = super().clone()
        if self.cloned_target is not None and self.path_to_root() == self.cloned_target:
            self.cloned_node = result

        return result


class UnaryExpression(MathExpression):
    """An expression that operates on one sub-expression"""

    def __init__(self, child: MathExpression = None, child_on_left: bool = True):
        super().__init__()
        self.child = child
        self.left_child = child_on_left
        self.set_child(child)

    def set_child(self, child: MathExpression = None) -> MathExpression:
        if self.left_child:
            return self.set_left(child)  # type:ignore
        else:
            return self.set_right(child)  # type:ignore

    def get_child(self) -> Optional[MathExpression]:
        if self.left_child:
            return self.left
        else:
            return self.right

    def evaluate(self, context: Dict[str, NumberType] = None) -> float:
        child = self.get_child()
        if child is None:
            raise ValueError("cannot evaluate unary expression without a valid child")
        return self.operate(child.evaluate(context))

    def operate(self, value: NumberType) -> NumberType:
        raise NotImplementedError("Must be implemented in subclass")


# ### Negation


class NegateExpression(UnaryExpression):
    """Negate an expression, e.g. `4` becomes `-4`"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["negate"]

    @property
    def name(self) -> str:
        return "-"

    def operate(self, value: NumberType) -> NumberType:
        return -value

    def __str__(self) -> str:
        return self.with_color("-{}".format(self.get_child()))

    def to_math_ml_fragment(self) -> str:
        """Convert this single node into MathML."""
        return f"-{super().to_math_ml_fragment()}"


# ### Function


class FunctionExpression(UnaryExpression):
    """A Specialized UnaryExpression that is used for functions.  The function name in
    text (used by the parser and tokenizer) is derived from the name() method on the
    class."""

    @property
    def name(self) -> str:
        raise NotImplementedError(
            "Must be implemented in subclass. Function is an abstract node"
        )

    def __str__(self) -> str:
        child = self.get_child()
        output = self.name
        if child:
            output = "{}({})".format(self.name, child)
        return self.with_color(output)


# ## Binary Expressions


class BinaryExpression(MathExpression):
    """An expression that operates on two sub-expressions"""

    def __init__(
        self,
        left: Optional[MathExpression] = None,
        right: Optional[MathExpression] = None,
    ):
        super().__init__(left=left, right=right)

    def evaluate(self, context: Dict[str, NumberType] = None) -> NumberType:
        left, right = self._check()
        return self.operate(left.evaluate(context), right.evaluate(context))

    @property
    def name(self) -> str:
        raise NotImplementedError("Must be implemented in subclass")

    def get_ml_name(self) -> str:
        return self.name

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        raise NotImplementedError("Must be implemented in subclass")

    def _check(self) -> Tuple[MathExpression, MathExpression]:
        if self.left is None or self.right is None:
            raise ValueError(
                "{}: left/right children must both be valid".format(
                    self.__class__.__name__
                )
            )
        return self.left, self.right

    def get_priority(self) -> int:
        """Return a number representing the order of operations priority
        of this node.  This can be used to check if a node is `locked`
        with respect to another node, i.e. the other node must be resolved
        first during evaluation because of it's priority.
        """
        priority = OOO_INVALID
        if isinstance(self, EqualExpression):
            priority = OOO_INVALID

        if isinstance(self, AddExpression) or isinstance(self, SubtractExpression):
            priority = OOO_ADDSUB

        if isinstance(self, MultiplyExpression) or isinstance(self, DivideExpression):
            priority = OOO_MULTDIV

        if isinstance(self, PowerExpression):
            priority = OOO_EXPONENT

        if isinstance(self, FunctionExpression):
            priority = OOO_FUNCTION

        return priority

    def self_parens(self) -> bool:
        self._check()
        """Return a boolean indicating whether this node should render itself with
        a set of enclosing parnetheses or not. This is used when serializing an
        expression, to ensure the tree maintains the proper order of operations. """
        binary_parent = cast(BinaryExpression, self.parent)
        parent_is_binary = isinstance(binary_parent, BinaryExpression)
        self_pri = self.get_priority()
        if not parent_is_binary or not binary_parent:
            return False

        parent_pri = binary_parent.get_priority()
        if parent_pri > self_pri:
            return True

        parent_side = binary_parent.get_side(self)
        # If we're a left child of a parent
        if parent_pri == self_pri:
            self_addsub = isinstance(self, (AddExpression, SubtractExpression))
            parent_addsub = isinstance(self.parent, (AddExpression, SubtractExpression))
            self_muldiv = isinstance(self, (MultiplyExpression, DivideExpression))
            parent_muldiv = isinstance(
                self.parent, (MultiplyExpression, DivideExpression)
            )
            if parent_side == "right" and self_addsub and parent_addsub:
                return True
            if parent_side == "left" and self_muldiv and parent_muldiv:
                return True
        return False

    def __str__(self) -> str:
        self._check()
        out = f"{self.left} {self.with_color(self.name)} {self.right}"
        return f"({out})" if self.self_parens() else out

    def to_math_ml_fragment(self) -> str:
        """Render this node as a MathML element fragment"""
        right, left = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        op_ml = self.make_ml_tag("mo", self.get_ml_name())
        if self.self_parens():
            return self.make_ml_tag(
                "mrow",
                "<mo>(</mo>{}{}{}<mo>)</mo>".format(left_ml, op_ml, right_ml),
                self.classes,
            )
        return self.make_ml_tag(
            "mrow", "{}{}{}".format(left_ml, op_ml, right_ml), self.classes
        )


class EqualExpression(BinaryExpression):
    """Evaluate equality of two expressions"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["equal"]

    @property
    def name(self) -> str:
        return "="

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        """This is where assignment of context variables might make sense. But context
        is not present in the expression's `operate` method.

        !!! warning

            TODO: Investigate this thoroughly.
        """
        raise ValueError("Unsupported operation. Euqality has no operation to perform.")


class AddExpression(BinaryExpression):
    """Add one and two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["add"]

    @property
    def name(self) -> str:
        return "+"

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        return one + two


class SubtractExpression(BinaryExpression):
    """Subtract one from two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["subtract"]

    @property
    def name(self) -> str:
        return "-"

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        return one - two


class MultiplyExpression(BinaryExpression):
    """Multiply one and two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["multiply"]

    @property
    def name(self) -> str:
        return "*"

    def get_ml_name(self) -> str:
        return "&#183;"

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        return one * two

    def __str__(self) -> str:
        """Multiplication special cases constant*variable to output `4x` instead of
        `4 * x`"""
        if isinstance(self.left, ConstantExpression):
            # const * var
            one = isinstance(self.right, VariableExpression)
            # const * var^power
            two = isinstance(self.right, PowerExpression) and isinstance(
                self.right.left, VariableExpression
            )
            if one or two:
                return self.with_color("{}{}".format(self.left, self.right))

        return super().__str__()

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        if isinstance(self.left, ConstantExpression):
            if isinstance(self.right, VariableExpression) or isinstance(
                self.right, PowerExpression
            ):
                return "{}{}".format(left_ml, right_ml)

        return super().to_math_ml_fragment()


class DivideExpression(BinaryExpression):
    """Divide one by two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["divide"]

    @property
    def name(self) -> str:
        return "/"

    def get_ml_name(self) -> str:
        return "&#247;"

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        left_ml = left.to_math_ml_fragment()
        right_ml = right.to_math_ml_fragment()
        return f"<mfrac><mi>{left_ml}</mi><mi>{right_ml}</mi></mfrac>"

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        if two == 0:
            return float("nan")
        else:
            return one / two


class PowerExpression(BinaryExpression):
    """Raise one to the power of two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["power"]

    @property
    def name(self) -> str:
        return "^"

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        # if left is mult, enclose only right in msup
        if isinstance(self.left, MultiplyExpression):
            left_ml = self.make_ml_tag("mrow", left_ml, self.classes)

        return self.make_ml_tag("msup", "{}{}".format(left_ml, right_ml), self.classes)

    def operate(self, one: NumberType, two: NumberType) -> NumberType:
        return np.power(one, two)

    def __str__(self) -> str:
        return "{}{}{}".format(self.left, self.with_color(self.name), self.right)


class ConstantExpression(MathExpression):
    """A Constant value node, where the value is accessible as `node.value`"""

    value: Optional[NumberType]

    @property
    def name(self) -> str:
        if self.value is not None and self.value % 1 == 0:
            return f"{int(self.value)}"
        return np.format_float_positional(self.value, trim="-")

    @property
    def type_id(self) -> int:
        return MathTypeKeys["constant"]

    def __init__(self, value: Optional[NumberType] = None):
        super().__init__()
        self.value = value

    def clone(self) -> "ConstantExpression":  # type:ignore[override]
        result = cast(ConstantExpression, super().clone())
        result.value = self.value
        return result  # type:ignore

    def evaluate(self, context: Dict[str, NumberType] = None) -> NumberType:
        assert self.value is not None
        return self.value

    def __str__(self) -> str:
        return self.with_color(self.name)

    def to_math_ml_fragment(self) -> str:
        return self.make_ml_tag("mn", str(self.value), self.classes)


class VariableExpression(MathExpression):
    identifier: Optional[str]

    @property
    def name(self) -> str:
        return f"{self.identifier}"

    @property
    def type_id(self) -> int:
        id = f"_{self.identifier.lower()[0]}" if self.identifier is not None else ""
        return MathTypeKeys[f"variable{id}"]

    def __init__(self, identifier: str = None):
        super().__init__()
        self.identifier = identifier

    def clone(self) -> "VariableExpression":  # type:ignore[override]
        result = cast(VariableExpression, super().clone())
        result.identifier = self.identifier
        return result

    def _check(self) -> None:
        if self.identifier is None:
            raise ValueError("identifier must be a letter")

    def __str__(self) -> str:
        self._check()
        return self.with_color("{}".format(self.identifier))

    def to_math_ml_fragment(self) -> str:
        self._check()
        return self.make_ml_tag("mi", str(self.identifier), self.classes)

    def evaluate(self, context: Dict[str, NumberType] = None) -> NumberType:
        self._check()
        id = cast(str, self.identifier)
        if context and context.get(id, None) is not None:
            return context[id]

        raise ValueError(
            "cannot evaluate statement with None variable: {}".format(self.identifier)
        )


class AbsExpression(FunctionExpression):
    """Evaluates the absolute value of an expression."""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["abs"]

    @property
    def name(self) -> str:
        return "abs"

    def operate(self, value: NumberType) -> NumberType:
        return np.absolute(value)


class SgnExpression(FunctionExpression):
    @property
    def type_id(self) -> int:
        return MathTypeKeys["sgn"]

    @property
    def name(self) -> str:
        return "sgn"

    def operate(self, value: NumberType) -> NumberType:
        """Determine the sign of an value.

        # Returns
        (int): -1 if negative, 1 if positive, 0 if 0"""
        if value < 0:
            return -1

        if value > 0:
            return 1

        return 0



from mathy_core.parser import ExpressionParser
from mathy_core.rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
    RestateSubtractionRule,
    VariableMultiplyRule,
)
from mathy_core.testing import run_rule_tests
from mathy_core.util import get_terms, terms_are_like


def test_rules_associative_property():
    def debug(ex):
        pass

    run_rule_tests("associative_property", AssociativeSwapRule, debug)


def test_rules_commutative_property():
    def debug(ex):
        pass

    run_rule_tests("commutative_property", CommutativeSwapRule, debug)


def test_rules_constants_simplify():
    def debug(ex):
        pass

    run_rule_tests("constants_simplify", ConstantsSimplifyRule, debug)


def test_rules_distributive_factor_out():
    def debug(ex):
        pass

    run_rule_tests("distributive_factor_out", DistributiveFactorOutRule, debug)


def test_rules_distributive_multiply_across():
    def debug(ex):
        pass

    run_rule_tests("distributive_multiply_across", DistributiveMultiplyRule, debug)


def test_rules_restate_subtraction():
    def debug(ex):
        pass

    run_rule_tests("restate_subtraction", RestateSubtractionRule, debug)


def test_rules_variable_multiply():
    def debug(ex):
        pass

    run_rule_tests("variable_multiply", VariableMultiplyRule, debug)


def test_rules_rule_can_apply_to():
    parser = ExpressionParser()
    expression = parser.parse("7 + 4x - 2")

    available_actions = [
        CommutativeSwapRule(),
        DistributiveFactorOutRule(),
        DistributiveMultiplyRule(),
        AssociativeSwapRule(),
    ]
    for action in available_actions:
        assert type(action.can_apply_to(expression)) == bool


def test_rules_like_terms_compare():
    parser = ExpressionParser()
    expr = parser.parse("10 + (7x + 6x)")
    terms = get_terms(expr)
    assert len(terms) == 3
    assert not terms_are_like(terms[0], terms[1])
    assert terms_are_like(terms[1], terms[2])

    expr = parser.parse("10 + 7x + 6")
    terms = get_terms(expr)
    assert len(terms) == 3
    assert not terms_are_like(terms[0], terms[1])
    assert terms_are_like(terms[0], terms[2])

    expr = parser.parse("6x + 6 * 5")
    terms = get_terms(expr)
    assert len(terms) == 2
    assert not terms_are_like(terms[0], terms[1])

    expr = parser.parse("360y^1")
    terms = get_terms(expr)
    assert len(terms) == 1

    expr = parser.parse("4z")
    terms = get_terms(expr)
    assert len(terms) == 1
